/*
 * Defini un graphe dont chaque noeud représente une arete
 * d'un autre graphe
 *
 *      Dominique Bouthinon
 *      LIPN Universite Paris 13
 */

#ifndef SRC_EDGES_GRAPH_HPP_
#define SRC_EDGES_GRAPH_HPP_

#include <functional>

#include "set_vertices_graph.hpp"

using namespace std ;

class edges_graph : public set_vertices_graph
{
	private :
												// correspondance entre les arcs x-y du graph initial
												// et le noeud a du graphe courant
												// ex) l'element <<2,3>, 5> signifie que l'arc 2-3
												// du graphe initial est represente par le noeud 5 du graphe des aretes
	map<pair<Tvertice_identifier, Tvertice_identifier>, Tvertice_identifier> edge_vertice_map ;

	void compute_edges_vertices() ;

	template <typename T>
	static void  compute_edges(
						    	graph<T>& g,
						        Tvertices_map<Tvertices_set>& communauty_vertices,
						        map<Tvertice_identifier, Tvertices_set>& map_vertice_edges_idents,
								map<pair<Tvertice_identifier, Tvertice_identifier>, Tvertice_identifier>& edge_vertice_map
					           ) ;

	static void compute_edges_adjacences(
									  	  Tvertices_map<Tvertices_set>& communauty_vertices,
										  map<Tvertice_identifier, Tvertice_identifiers_set >& map_vertice_communauty_idents,
										  Tadjacences_map& adjacences
									     ) ;

				// verifie que l'arete x et l'arete y font partie
				// d'un triangle du graphe g
	template <typename T>
	bool dans_triangle(
				  	    const Tvertice_identifier& x,
						const Tvertice_identifier& y,
						graph<T>& g
				 	  ) ;

	template <typename T>
	bool dans_k_clique(
			  	  	  	const Tvertice_identifier& a,
						const Tvertice_identifier& b,
						const int k,
						graph<T>& g,
						set<pair<Tvertice_identifier, Tvertice_identifier> >& edges_ok
			 	 	  ) ;

	bool base_k_couronne(
							Tvertice_identifier& u,
							Tvertice_identifier& v,
							Tvertice_identifiers_set& adj_u,
							Tvertice_identifiers_set& adj_v,
							const int k,
							set<pair<Tvertice_identifier, Tvertice_identifier> >& edges_ok
						) ;

	template <typename T>
	bool dans_k_couronne(
			  	  	  	 const Tvertice_identifier& a,
						 const Tvertice_identifier& b,
						 const int k, 	// k >= 3
						 graph<T>& g,
						 set<pair<Tvertice_identifier, Tvertice_identifier> >& edges_ok
			 	 	  ) ;

	public:

		// --------------------- CONSTRUCTEURS ---------------------------------

		edges_graph() ;

		edges_graph(
						string aName,
						Tvertices_map<Tvertices_set>& vertices,  // noeud : <id_communaute, {noeud1, noeud2}>
						Tadjacences_map& adjacences,
						function<void (Tvertices_set)> display_vertice_value_function,
						function<string (Tvertices_set)> toString_vertice_value_function,
						map<Tvertice_identifier, Tvertice_identifiers_set>& map_vertice_communauties_idents,
						map<pair<Tvertice_identifier, Tvertice_identifier>, Tvertice_identifier>& edge_vertice_map
				 	 	) ;

		edges_graph(
						string aName,
						Tvertices_map<Tvertices_set>& vertices,
						function<void (Tvertices_set)> display_vertice_value_function,
						function<string (Tvertices_set)> toString_vertice_value_function
					   ) ;

					// constructeur par copie
		edges_graph(const edges_graph& g) ;


	template <typename T>
	static edges_graph*  compute_edges_graph(graph<T>& g) ;


	// ----------------- ABSTRACTIONS D'ARETES --------------------------------

		// conserve les aretes a-b tel que a-b est dans un triangle
	template <typename T>
	void keep_triangles(graph<T>& g) ;

		// conserve les aretes a-b du graphe des aretes courant dérivé de G
		// tel que les sommets de a-b sont dans une k-clique de G
	template <typename T>
	void keep_k_cliques_ou_k_couronnes(const int k_clique, const int k_couronne, graph<T>& g) ;

} ;


// =========================== IMPLEMENTATIONS =================================

// ----------------------------- CONSTRUCTEURS ---------------------------------

edges_graph::
edges_graph() : set_vertices_graph()
{
}

	 // construit le graphe a partir d'une liste de noeuds et
	 // d'adjacences (avec predecesseurs et successeurs).
	 // ATTENTION : les objets internes sont des copies des paramètres
edges_graph::
edges_graph(
				string aName,
				Tvertices_map<Tvertices_set>& vertices,
				Tadjacences_map& adjacences,
				function<void (Tvertices_set)> display_vertice_value_function,
				function<string (Tvertices_set)> toString_vertice_value_function,
				map<Tvertice_identifier, Tvertice_identifiers_set>& map_vertice_communauties_idents,
				map<pair<Tvertice_identifier, Tvertice_identifier>, Tvertice_identifier>& edge_vertice_map
				)
				: set_vertices_graph(aName, vertices, adjacences,
						           display_vertice_value_function,
								   toString_vertice_value_function,
								   map_vertice_communauties_idents)
{
	this->edge_vertice_map = edge_vertice_map ;
}

edges_graph::
edges_graph(
				string aName,
				Tvertices_map<Tvertices_set>& vertices,
				function<void (Tvertices_set)> display_vertice_value_function,
				function<string (Tvertices_set)> toString_vertice_value_function
				)
				: set_vertices_graph(aName, vertices,
						             display_vertice_value_function,
									 toString_vertice_value_function)
{
}

		// constructeur par copie
edges_graph::
edges_graph(const edges_graph& g) : set_vertices_graph(g)
{

}


// créé le graphe ge des arcs dérivé du graphe courant
 // un arc de g devient un noeud dans ge
// !!!!! ATTENTION A L'ORIENTATION !!! !!
template <typename T>
edges_graph* edges_graph::compute_edges_graph(graph<T>& g)

{
		Tvertices_map<Tvertices_set> communauty_vertices ;
		map<Tvertice_identifier, Tvertices_set> map_vertice_edges_idents ;
		map<pair<Tvertice_identifier, Tvertice_identifier>, Tvertice_identifier> edge_vertice_map ;

		compute_edges(g, communauty_vertices, map_vertice_edges_idents, edge_vertice_map) ;

								// creation des adjacences entre communautes
		Tadjacences_map adjacences ;

		compute_edges_adjacences(communauty_vertices,
								 map_vertice_edges_idents,
								 adjacences) ;

		return new edges_graph(
								"edges graph",
								communauty_vertices,
								adjacences,
								graph<Tvertices_set>::display_vertice_identifiers_set,
								graph<Tvertices_set>::toString_vertice_identifiers_set,
								map_vertice_edges_idents,
								edge_vertice_map
							   ) ;
	}


	/**
	 * retourne la liste d'adjacences correspondant aux 2 communautés du graphe courant
	 */
void edges_graph::compute_edges_adjacences(
							 Tvertices_map<Tvertices_set>& communauty_vertices,
							 map<Tvertice_identifier, Tvertice_identifiers_set >& map_vertice_communauty_idents,
							 Tadjacences_map& adjacences
							)
{
		Tvertices_map_iterator<Tvertices_set> it 	 = communauty_vertices.begin() ;
		Tvertices_map_iterator<Tvertices_set> it_end = communauty_vertices.end() ;

															// pour chaque noeud representant un arc
		for ( ; it != it_end ; it++)
		{
				const Tvertice_identifier& edge_ident = it->first ;
															// it-> first est un numéro d'arc

				Tvertices_set& apexs = it->second.value ;		// ensemble des sommets de l'arc

															// 2 sommets de l'arc
				const Tvertice_identifier& apex1    = *(apexs.begin()) ;
				const Tvertice_identifier& apex2    = *(++(apexs.begin())) ;

													// on recupere l'ensemble des arcs
													// comprenant ces sommets
				Tvertice_identifiers_set& t_set1  = (map_vertice_communauty_idents)[apex1] ;
				Tvertice_identifiers_set& t_set2  = (map_vertice_communauty_idents)[apex2] ;

													// contriendra les arcs adjacents à l'arc edge_ident
				Tvertice_identifiers_set edges_adjacents ;

				set_union(t_set1.begin(),t_set1.end(), t_set2.begin(),t_set2.end(),
						  inserter(edges_adjacents, edges_adjacents.begin()));

				edges_adjacents.erase(edge_ident) ;		// l'arc edge_ident n'est pas adjacent à lui-même

				Tadjacences adja ;
															// répartit les edges adjacents
															// en prédecesseurs ou successeurs
				Tvertice_identifiers_set_iterator it_t     = edges_adjacents.begin() ;
				Tvertice_identifiers_set_iterator it_t_end = edges_adjacents.end() ;

				for (; it_t != it_t_end ; it_t++)
				{
						if (*it_t < edge_ident)
							adja.predeccessors.insert(*it_t) ;
						else
							adja.successors.insert(*it_t) ;
				}

				adjacences.insert(make_pair(edge_ident, adja)) ;
		}
	}


	/**
	 * retourne les 2 communauties (ensemble des noeuds formant des arcs
	 * dans le graph g)
	 * ATTENTION ON TIENT COMPTE DE L'ORIENTATION INITIALE LORS DE LA CREATION
	 * DES DONNEES MINERLC
	 */
	template <typename T>
	void  edges_graph::compute_edges(
						graph<T>& g,
						Tvertices_map<Tvertices_set>& communauty_vertices,
						map<Tvertice_identifier, Tvertices_set>& map_vertice_edges_idents,
						map<pair<Tvertice_identifier, Tvertice_identifier>, Tvertice_identifier>& edge_vertice_map
				                     )
	{
		Tvertice_identifier c_ident = 0 ;			// identifiant d'un futur noeud
													// du graphe des aretes

		Tadjacences_map_iterator x      = g.get_adjacences_map().begin() ;
		Tadjacences_map_iterator x_end  = g.get_adjacences_map().end() ;

													// pour chaque noeud du graphe initial
		for (; x != x_end ; x++)
		{
				Tvertice_identifiers_set & successors = (*x).second.successors ;

				Tvertice_identifiers_set_iterator y_end = successors.end() ; // fin list des successeurs

													// pour chacun de ses successeurs
				for (Tvertice_identifiers_set_iterator y = successors.begin() ; y != y_end ; y++)
				{
							Tvertice<Tvertices_set> vertice ;

							Tvertices_set s = {x->first, *y} ;

							vertice.value     = s ;
							vertice.label     = "" ;
							vertice.component = -1 ;

												// forme un noeud representant l'arete (x-y)
							communauty_vertices.insert(make_pair(c_ident, vertice));

												// on etablie les liens entre les 2 noeuds
												// et le nouvel arc dont ils sont sommets
							(map_vertice_edges_idents)[x->first].insert(c_ident) ;
							(map_vertice_edges_idents)[*y].insert(c_ident) ;

												// sauvegarde que l'arete x-y du graphe initial correspond au noeud c_ident
												// on sait que x < y (en fait *x < *y)

							// cout << "x = " << x->first << endl ;
							// cout << "y = " << *y << endl ;
							// cout << "ident = " << c_ident << endl ;

							edge_vertice_map.insert(make_pair(make_pair(x->first, *y), c_ident)) ;

							c_ident++ ;
				}
				// fin for it1
		}
		// fin for it
	}
	// fin fonction compute_2_communauties


// =========== CALCULE LES K-COMMUNAUTES de cliques ou de couronnes liées par au moins une arete =========

		// on verifie que (u-v) est la base d'une
		// éventuelle k-couronnes :
		// on verifie qu'ils ont k-2 voisins communs
bool edges_graph::base_k_couronne(
									Tvertice_identifier& u,
									Tvertice_identifier& v,
									Tvertice_identifiers_set& adj_u,
									Tvertice_identifiers_set& adj_v,
									const int k,
									set<pair<Tvertice_identifier, Tvertice_identifier> >& edges_ok
								 )
{
	Tvertice_identifiers_set adj_uv ;
	set_intersection (
						adj_u.begin(), adj_u.end(),
						adj_v.begin(), adj_v.end(),
						inserter(adj_uv,adj_uv.begin())
					 ) ;

	// cout << "k = " << k << endl ;

	// cout << "u = " << u << endl ;
	// cout << "v = " << v << endl ;

	// displaySet("adj_u", adj_u) ;
	// displaySet("adj_v", adj_v) ;
	// displaySet("adj_uv",adj_uv) ;


	if (adj_uv.size() < k-2)
		return false ;						// pas de k-courronnes de base u-v

											// on recupere l'arete (u-v) dans le graphe des aretes

	// recupere le noeud du graphe des aretes correspondant à l'arete u-v
	Tvertice_identifier uv ;

	if (u < v)
		uv = this->edge_vertice_map[make_pair(u,v)] ;
	else
		uv = this->edge_vertice_map[make_pair(v,u)] ;

										// tous les sommets w adjacents à u et v sont des
										// aretes (u-w) et (v-w) du graphe des aretes telles que
										// (u-v)-(u-w), (u-v)-(v-w) et (u-w)-(v-w)
										// il faut donc ajouter ces 3 aretes dans le graphe des aretes

	for (Tvertice_identifiers_set_iterator w = adj_uv.begin() ;
		  w!= adj_uv.end() ; w++)
	{
			// recupere le noeud du graphe des aretes correspondant à l'arete u-w
		Tvertice_identifier uw ;

		if (u < *w)
			uw = this->edge_vertice_map[make_pair(u,*w)] ;
		else
			uw = this->edge_vertice_map[make_pair(*w,u)] ;

		Tvertice_identifier vw ;

		if (v < *w)
			vw = this->edge_vertice_map[make_pair(v,*w)] ;
		else
			vw = this->edge_vertice_map[make_pair(*w,v)] ;

		edges_ok.insert(make_pair(uv, uw)) ;
		edges_ok.insert(make_pair(uv, vw)) ;
		edges_ok.insert(make_pair(uw, vw)) ;
	}

	return true ;
}


// verifie si l'arete a-b du graphe des aretes, avec a = {y,x}, b = {x,z}, est telle que x, y et z appartiennent
// à une k-couronnes de g
// un k-couronnes est formé d'une base u-v (u et v sont des sommets de g), tel que
// u et v ont (au moins) k-2 voisins communs.
template <typename T>
bool edges_graph::dans_k_couronne(
		  	  	  	  	  	  	  	  const Tvertice_identifier& a,
									  const Tvertice_identifier& b,
									  const int k, 	// k >= 4
									  graph<T>& g,
									  set<pair<Tvertice_identifier, Tvertice_identifier> >& edges_ok
		 	 	 	 	 	     )
{
	Tvertice_identifiers_set va = this->get_vertice_value(a) ; // arete {x,y} de g
	Tvertice_identifiers_set vb = this->get_vertice_value(b) ; // arete {x,z} de g

	// displaySet("va", va) ;
	// displaySet("vb", vb) ;

	Tvertice_identifiers_set vc ;
											 // vc = va (difference symetrique) vb = vc = {y,z}

	set_symmetric_difference(va.begin(), va.end(), vb.begin(), vb.end(),
							inserter(vc, vc.begin())) ;

	// displaySet("vc", vc) ;
											// on recupere les 2 sommets de vc
	Tvertice_identifiers_set_iterator it = vc.begin() ;
	Tvertice_identifier y = *it ;

	//cout << "y = " << y  << endl ;
	it++ ;

	Tvertice_identifier z = *it ;

	//cout << "z = " << z << endl ;
											// verifie que y et z sont adjacents dans g
	if (!g.adjacent(y, z))
		return false ;
								 // y et z sont adjacents
								 // => x,y,z forment un triangle
								 // chaque couple a au moins un voisin commun

								 // on recupere le noeud commun : x
   // cout << "adjacent"  << endl ;

	Tvertice_identifiers_set vx ;
	set_intersection (
					  va.begin(), va.end(),
					  vb.begin(), vb.end(),
					  inserter(vx,vx.begin())
					 ) ;

	// displaySet("vx", vx) ;

	Tvertice_identifier x = *(vx.begin()) ;
											// recupere les noeuds adjacents
											// à x, y et z
	//cout << "x = " << x << endl ;

	Tvertice_identifiers_set adj_x ;
	g.adjacent_vertices(x, adj_x) ;
	// displaySet("adj_x", adj_x) ;

	Tvertice_identifiers_set adj_y ;
	g.adjacent_vertices(y, adj_y) ;
	// displaySet("adj_y", adj_y) ;

	Tvertice_identifiers_set adj_z ;
	g.adjacent_vertices(z, adj_z) ;
	// displaySet("adj_z", adj_z) ;
											// teste est si (x-y) est la base
											// d'une (k-1)-couronne

	if (base_k_couronne(x, y, adj_x, adj_y, k, edges_ok))
		return true ;
											// teste est si (x-z) est la base
											// d'une k-couronne

	if (base_k_couronne(x, z, adj_x, adj_z, k, edges_ok))
		return true ;

											// teste est si (y-z) est la base
											// d'une k-couronne

	if (base_k_couronne(y, z, adj_y, adj_z, k, edges_ok))
		return true ;

	return false ;
}



// verifie si les sommets x et y (aretes de g) font partie d'un triangle de g
	template <typename T>
	bool edges_graph::dans_triangle(
			  	  	  	  	  	  	  const Tvertice_identifier& a,
									  const Tvertice_identifier& b,
									  graph<T>& g
			 	 	 	 	 	    )
	{
		Tvertice_identifiers_set vx = this->get_vertice_value(a) ; // arete {x,y} de g
		Tvertice_identifiers_set vy = this->get_vertice_value(b) ; // arete {x,z} de g

											// Attention on ne connait pas l'ordre
											// des elements dans vx et vy
		// displaySet("vx", vx) ;
		// displaySet("vy", vy) ;
														// vx difference symetrique vy = {y,z}
		Tvertice_identifiers_set vz ;

		set_symmetric_difference(vx.begin(), vx.end(), vy.begin(), vy.end(),
								inserter(vz, vz.begin())) ;

		// displaySet("vz", vz) ;

														// on recupere les 2 sommets de vz
		Tvertice_identifiers_set_iterator it = vz.begin() ;
		Tvertice_identifier u = *it ;

		// cout << "u = " << u  << endl ;
		it++ ;

		Tvertice_identifier v = *it ;

		// cout << "v = " << v << endl ;
														// verifie que u et v sont adjacents dans g
		return g.adjacent(u, v) ;
	}



// verifie si l'arete a-b du graphe des aretes, avec a = {x,y}, b = {x,z}, tel que x, y et z appartiennent
// à une k-clique de g
template <typename T>
bool edges_graph::dans_k_clique(
		  	  	  	  	  	  	  const Tvertice_identifier& a,
								  const Tvertice_identifier& b,
								  const int k, 	// k >= 3
								  graph<T>& g,
								  set<pair<Tvertice_identifier, Tvertice_identifier> >& edges_ok
		 	 	 	 	 	    )
{
	Tvertice_identifiers_set va = this->get_vertice_value(a) ; // arete {x,y} de g
	Tvertice_identifiers_set vb = this->get_vertice_value(b) ; // arete {x,z} de g

												// ???? quel est l'ordre de x, y et z dans le graph initial ???????

	// displaySet("va", va) ;
	// displaySet("vb", vb) ;
													// va difference symetrique vb = {y,z}
													// noeuds differents
	Tvertice_identifiers_set vc ;

	set_symmetric_difference(va.begin(), va.end(), vb.begin(), vb.end(),
							inserter(vc, vc.begin())) ;

	// displaySet("vc", vc) ;
											// on recupere les 2 sommets de vc
	Tvertice_identifiers_set_iterator it = vc.begin() ;
	Tvertice_identifier y = *it ;

	// cout << "y = " << y  << endl ;
	it++ ;

	Tvertice_identifier z = *it ;

	// cout << "z = " << z << endl ;
											// verifie que y et z sont adjacents dans g
	if (!g.adjacent(y, z))
		return false ;
											// y et z sont adjacents

											// recupere le noeud commun : x
	Tvertice_identifiers_set vx ;
	set_intersection (
			             va.begin(), va.end(),
						 vb.begin(), vb.end(),
			             inserter(vx,vx.begin())
					    ) ;

	// displaySet("vx", vx) ;

	Tvertice_identifier x = *(vx.begin()) ;
											// recupere les noeuds adjacents
											// à x, y et z
	Tvertice_identifiers_set adj_x ;
	g.adjacent_vertices(x, adj_x) ;
	// displaySet("adj_x", adj_x) ;

	Tvertice_identifiers_set adj_y ;
	g.adjacent_vertices(y, adj_y) ;
	// displaySet("adj_y", adj_y) ;

	Tvertice_identifiers_set adj_z ;
	g.adjacent_vertices(z, adj_z) ;
	// displaySet("adj_z", adj_z) ;

	Tvertice_identifiers_set adj_xy ;
	set_intersection (
					   adj_x.begin(), adj_x.end(),
					   adj_y.begin(), adj_y.end(),
					   inserter(adj_xy,adj_xy.begin())
					  ) ;

	Tvertice_identifiers_set adj_xyz ;
	set_intersection (
				             adj_xy.begin(), adj_xy.end(),
							 adj_z.begin(), adj_z.end(),
				             inserter(adj_xyz,adj_xyz.begin())
				     ) ;

	// displaySet("adj_xyz", adj_xyz) ;

	Tvertice_identifiers_set xyz ;			// sommets impliques ;

	set_union(va.begin(), va.end(), vb.begin(), vb.end(),
			  inserter(xyz, xyz.begin())) ;

											// retire les noeuds {x,y,z} parmi
											// les noeuds ajdacents à x, et y et z

	Tvertice_identifiers_set adja ;
	set_difference(adj_xyz.begin(), adj_xyz.end(), xyz.begin(), xyz.end(),
	               inserter(adja, adja.begin()));

	// displaySet("adja", adja) ;

											// recherche d'une k-3 clique dans adja
	if (adja.size() < k-3)
		return false ;						// pas de k-3 clique dans adja

	Tvertice_identifiers_set clique ;

	// cout << "x = " << x << endl ;
	// cout << "y = " << y << endl ;
	// cout << "z = " << z << endl ;


	if (g.existe_k_clique(k-3, adja, clique))
	{
						// ajoute les noeuds x,y,z à clique
		Tvertice_identifiers_set clique_finale ;

		// displaySet("xyz",xyz) ;
		// displaySet("clique", clique) ;

		set_union(clique.begin(), clique.end(), xyz.begin(), xyz.end(),
				  inserter(clique_finale, clique_finale.begin()))  ;

							// clique_finale est une k-clique

		// displaySet("clique finale", clique_finale) ;

						// !!! ATTENTION LA CLIQUE FINALE EST (NORMALEMENT) ORDONNEE!!!
						// ON UTILISE CET ORDRE
						// reconstitue les aretes du graphe des aretes en fonction
						// de la k-clique du graphe initiale

		for (Tvertice_identifiers_set_iterator x = clique_finale.begin() ;
			 x != clique_finale.end() ; x++)
		{
				for (Tvertice_identifiers_set_iterator y = next(x,1) ;
					 y != clique_finale.end() ; y++)
				{
									// on sait que x < y (en fait *x < *y)

									// recupere le noeud du graphe des aretes correspondant à l'arete x-y
					Tvertice_identifier a = this->edge_vertice_map[make_pair(*x,*y)] ;

									// l'arete a = (x-y) est adjacente à toutes les aretes :

					for (Tvertice_identifiers_set_iterator z = next(y,1) ;
						  z != clique_finale.end() ; z++)
					{
						// cout << endl << "*x = " << *x << endl ;
						// cout << "*y = " << *y << endl ;
						// cout << "*z = " << *z << endl << endl ;

									//  i) x-z avec z > y
									// recupere le noeud du graphe des arets correspondant à l'arete x-z
						Tvertice_identifier b1 = this->edge_vertice_map[make_pair(*x,*z)] ;

						// displayPair(make_pair(*x, *z)) ;
						// cout << endl ;
						// cout << "a1,b1  : " ;
						// displayPair(make_pair(a, b1)) ;
						// cout << endl ;

						// edges_ok.insert(make_pair(a, b)) ;
														// (x-y) adjacent à (x-z)
						edges_ok.insert(make_pair(a, b1)) ;

									//  ii) y-z avec z > y
									// recupere le noeud correspondant à l'arete x-z
						Tvertice_identifier b2 = this->edge_vertice_map[make_pair(*y,*z)] ;

						// displayPair(make_pair(*y, *z)) ;
						// cout << endl ;
						// cout << "a2,b2  : " ;
						// displayPair(make_pair(a, b2)) ;
						// cout << endl ;
														// (x-y) adjacent à (y-z)
						edges_ok.insert(make_pair(a, b2)) ;
					}
				}
		}
		return true ;
	}

	return false ;
}


	// on retire du graphe des aretes this , les aretes (a-b) de this qui ne sont pas
	// dans une k-clique ou une k-couronne du graphe g
	// dans le graphe resultant, chaque composante connexe
	// contient les k-cliques ou k-couronnes du graphe original liées deux à deux par une arete
template <typename T>
void edges_graph::keep_k_cliques_ou_k_couronnes(
							     const int k_clique,
								 const int k_couronne,
							     graph<T>& g
						        )
{
												// on sait que soit k_clique == 0 soit k_crown == 0
	if (k_clique == 3 || k_couronne == 3)
	{
		this->keep_triangles(g) ;
		return ;
	}
													// k >= 3
													// conserve les aretes satisfaisantes
	set<pair<Tvertice_identifier, Tvertice_identifier> > edges_ok ;

								// pour chaque noeud a (arete de g)
	for (Tvertice_identifiers_set_iterator a = this->get_vertice_identifiers_list().begin() ;
		 a != this->get_vertice_identifiers_list().end() ; a++)

	{
			// cout << "a = " << *a << endl ;

			Tvertice_identifiers_set& successors_a = this->get_successors(*a) ;
			Tvertice_identifiers_set_iterator b    = successors_a.begin() ;

											// pour chaque successeur b de a
			while (b != successors_a.end())
			{
				// cout << "b = " << *b << endl ;
												// on considere l'arete (a,b) du graph des aretes
												// on sait a < b (en fait *a < *b)

				pair<Tvertice_identifier, Tvertice_identifier> p = make_pair(*a,*b) ;

				// cout << endl ;
			    // cout << "arete : " << *a << "-" << *b << endl ;

				if (edges_ok.find(p) == edges_ok.end())
				{
					bool arete_ok = false ;
												// arete a-b pas encore vue
					if (k_clique > 0)
					{
						arete_ok = dans_k_clique(*a,*b, k_clique, g, edges_ok) ;
					}
					else if (k_couronne > 0)
					{
						arete_ok = dans_k_couronne(*a,*b, k_couronne, g, edges_ok) ;
					}

					if (arete_ok)
					{
							//cout << "ok" << endl ;
							// displaySetPair("edges_ok", edges_ok) ;
							b++ ;
					}
					else
					{
						// cout << "effacee" << endl ;
													// arete a-b non satisfaisante
													// efface l'arete (a,b)
						this->get_predeccessors(*b).erase(*a) ;  // efface a des predecesseurs de b
						b = successors_a.erase(b)    ;           // efface b des successeurs de a
					}
				}
				else								// arete a-b deja vue
				{
					// cout << "deja vu" << endl ;
					b++ ;
				}
		   }
			// fin pour chaque b successeur de a
	 }
	 // fin pour chaque a

	this->discard_isolated_vertices() ;
}
// fin methode



// on retire du graphe des aretes, les aretes qui ne sont pas dans un triangle
// du graphe g
// !!! ATTENTION !!!
// les données de minerLC sont ordonnées, le graphe g est donc ordonné
// par ces données :
// ex) si l'objet a précède l'objet b alors l'arc a->b est dans le graph
// mais l'arc b->a ne l'est pas
template <typename T>
void edges_graph::keep_triangles(graph<T>& g)
{
							// pour chaque noeud a (arete de g)
for (Tvertice_identifiers_set_iterator a = this->get_vertice_identifiers_list().begin() ;
	 a != this->get_vertice_identifiers_list().end() ; a++)

{
		// cout << "a = " << *a << endl ;

		Tvertice_identifiers_set& successors_a = this->get_successors(*a) ;
		Tvertice_identifiers_set_iterator b    = successors_a.begin() ;

										// pour chaque successeur b de a
		while (b != successors_a.end())
		{
			// cout << "b = " << *b << endl ;
											// on considere l'arete (a,b)
				if(dans_triangle(*a,*b, g))
				{
					b++ ;
				}
				else
				{
												// arete (a,b) non satisfaisante
												// efface l'arete (a,b)
					this->get_predeccessors(*b).erase(*a) ;  // efface a des predecesseurs de b
					b = successors_a.erase(b)    ;           // efface b des successeurs de a
				}
       }
        // pour chaque b successeur de a
 }
 // fin pour chaque a


this->discard_isolated_vertices() ;

}
// fin methode abstraction_3_cliques



#endif /* SRC_EDGES_GRAPH_HPP_ */
