/*
 * Defini un graphe dont chaque noeud est un ensemble de sommets
 * (une communauté) d'un autre graphe
 *
 *  Created on: 21 oct. 2015
 *      Author: dominique
 */

#ifndef SRC_SET_VERTICES_GRAPH_HPP_
#define SRC_SET_VERTICES_GRAPH_HPP_

#include <functional>

using namespace std ;

class set_vertices_graph : public graph<Tvertices_set>
{
	private :
								// liste  {<id_noeud_graph_initial, {id_communaute1, idcommunauté2}}
								// où id_noeud_graph_initial est dans les communautés (noeuds du graph courant)
								// identifiées par id_commuante1, etc...
		map<Tvertice_identifier, Tvertice_identifiers_set> map_vertice_communauties_idents ;

	public:

		// --------------------- CONSTRUCTEURS ---------------------------------

		set_vertices_graph() ;

		set_vertices_graph(
						string aName,
						Tvertices_map<Tvertices_set>& vertices,  // noeud : <id_communaute, {noeud1, noeud2,...}>
						Tadjacences_map& adjacences,
						function<void (Tvertices_set)> display_vertice_value_function,
						function<string (Tvertices_set)>  toString_value_function,
						map<Tvertice_identifier, Tvertice_identifiers_set>& map_vertice_communauties_idents
		 	 	 	   ) ;

		set_vertices_graph(
						string aName,
						Tvertices_map<Tvertices_set>& vertices,
						function<void (Tvertice_identifiers_set)> display_vertice_value_function,
						function<string (Tvertice_identifiers_set)>  toString_value_function
					   ) ;


					// constructeur par copie
		set_vertices_graph(const set_vertices_graph& g) ;


		virtual ~set_vertices_graph(){}


		//------------------------- GETTERS ------------------------------------

		map<Tvertice_identifier, Tvertice_identifiers_set> get_map_vertice_communauties_idents() ;

		// void get_map_vertice_communauties_idents() ;


		//------------------------- AFFICHAGE ----------------------------------

		void display_graph() ;

		// ----------------------- APPLATISSEMENT ------------------------------

		Tvertice_identifiers_set* flat_component(Tvertice_identifiers_set& component) ;

} ;



// =========================== IMPLEMENTATIONS =================================


// ----------------------------- CONSTRUCTEURS ---------------------------------

set_vertices_graph::
set_vertices_graph() : graph()
{
}

	 // construit le graphe a partir d'une liste de noeuds et
	 // d'adjacences (avec predecesseurs et successeurs).
	 // ATTENTION : les objets internes sont des copies des paramètres
set_vertices_graph::
set_vertices_graph(
				string aName,
				Tvertices_map<Tvertices_set>& vertices,
				Tadjacences_map& adjacences,
				function<void (Tvertices_set)> display_vertice_value_function,
				function<string (Tvertices_set)>  toString_value_function,
				map<Tvertice_identifier, Tvertice_identifiers_set>& map_vertice_communauties_idents
				)
				: graph(aName, vertices, adjacences,
						display_vertice_value_function,  toString_value_function)
{
	this->map_vertice_communauties_idents = map_vertice_communauties_idents ;
}


set_vertices_graph::
set_vertices_graph(
				string aName,
				Tvertices_map<Tvertices_set>& vertices,
				function<void (Tvertices_set)> display_vertice_value_function,
				function<string (Tvertices_set)>  toString_value_function
				)
				: graph(aName, vertices,
						display_vertice_value_function, toString_value_function)
{
}


		// constructeur par copie
set_vertices_graph::
set_vertices_graph(const set_vertices_graph& g) : graph(g)
{
}


//------------------------- GETTERS ------------------------------------
map<Tvertice_identifier, Tvertice_identifiers_set>
set_vertices_graph::get_map_vertice_communauties_idents()
{
	return this->map_vertice_communauties_idents ;
}


// ------------------------------- AFFICHAGE -----------------------------------

void set_vertices_graph::display_graph()
{
	graph::display_graph() ;

	map<Tvertice_identifier, Tvertice_identifiers_set>::iterator
					it = this->map_vertice_communauties_idents.begin() ;

	map<Tvertice_identifier, Tvertice_identifiers_set>::iterator
					it_end = this->map_vertice_communauties_idents.end() ;

	cout << "{vertice-communauties}: " << endl ;

	for ( ; it != it_end ; it++)
	{
		cout << it->first << " : " ;
		graph::display_vertice_identifiers_set(it->second) ;
		cout << endl ;
	}
}


// ------------------------------- APPLATISSEMENT ------------------------------

		// retourne la liste des identificateurs des sommets du graphe initial
		// impliqués dans les communautes de la composante connexe component
		// du graphe courant
Tvertice_identifiers_set* set_vertices_graph::
flat_component(Tvertice_identifiers_set& component)
{
	Tvertice_identifiers_set_iterator it     = component.begin() ;
	Tvertice_identifiers_set_iterator it_end = component.end() ;

	Tvertice_identifiers_set*  initial_objects  = new Tvertice_identifiers_set ;

			// pour chaque noeud de la composante connexe component
	for ( ; it != it_end ; it++)
	{
				// recupere les noeuds du graphe initial de la communauté *it
		Tvertice_identifiers_set& current_objects = this->get_vertice_value(*it) ;
		initial_objects->insert(current_objects.begin(), current_objects.end()) ;
	}

	return initial_objects ;
}


#endif /* SRC_SET_VERTICES_GRAPH_HPP_ */
