/* version minerLC */

/*
 * weighted_graphe.hpp
 *
 *  Created on: 9 sept. 2015
 *      Author: Dominique Bouthinon
 *
 *  Representation d'un graphe oriente par une liste d'adjacences
 *  les valeurs des noeuds sont de n'importe quel type
 *  represente par le type template Tvalue
 *
 */

#ifndef WEIGHTED_GRAPHE_HPP_
#define WEIGHTED_GRAPHE_HPP_

#include <functional>
#include <iostream>
#include <algorithm>
#include <exception>
#include <stack>

#include "graph.hpp"
#include "types_graph.hpp"
#include "abstraction.hpp"
#include "utils_divers.hpp"

using namespace std ;

/* =============================== class graph ============================== */

template <typename Tvalue>
class weighted_graphe : public virtual graph<Tvalue>
{
	private :

	Tweighted_edges_map weighted_edges_map ;

	public :

	// --------------------------- CONSTRUCTEURS -------------------------------


	weighted_graphe() ;				// graph sans information

			 	 	 	 	// construit le graphe avec un nom et une fonction d'affichage
	weighted_graphe(
					string aName,
					function<void (Tvalue)> display_vertice_value_function
		 	 	  ) ;


					 // construit le graphe a partir d'une liste de noeuds et
					 // d'adjacences (avec predecesseurs et successeurs).
					 // ATTENTION : les objets internes sont des copies des paramètres
	weighted_graphe(
					string aName,
					Tvertices_map<Tvalue>& vertices,
					Tadjacences_map& adjacences,
					Tweighted_edges_map& edges_map,
					function<void (Tvalue)> display_vertice_value_function
				   ) ;


	weighted_graphe(
					string aName,
					Tvertices_map<Tvalue>& vertices,
					Tadjacences_map& adjacences,
					Tweighted_edges_map& edges_map,
					function<void (Tvalue)> display_vertice_value_function,
					function<string (Tvalue)> toString_value_function
					) ;

	weighted_graphe(
					string aName,
					Tvertices_map<Tvalue>& vertices,
					function<void (Tvalue)> display_vertice_value_function
		 	 	 ) ;

					// idem a partir d'un graph g
	weighted_graphe(const weighted_graphe<Tvalue>& g) ;


	virtual ~weighted_graphe() ;


	graph<Tvalue>* copy_graph() ;

					// accesseurs
	Tweighted_edges_map& get_edges_map() ;

	// accesseurs
	Tweighted_edges_map get_copy_edges_map() ;

					// mutateurs
	void set_edges_map(Tweighted_edges_map& edges_map) ;

					// retourne le degre pondere du noeud ident
	float degree(Tvertice_identifier ident) ;

	void save_graph_dot(string name, bool vertice_ident, bool vertice_label) ;

	void save_graph_dot2(string graph_name,
						 bool vertice_ident,
						 bool vertice_label,
						 bool directed_graph) ;

} ;
// fin classe weighted_graphe<Tvalue>

/* ===================== IMPLEMENTATIONS FONCTIONS ========================== */

//------------------------------------ CONSTRUCTEURS ---------------------------

template <typename Tvalue>
weighted_graphe<Tvalue>::
weighted_graphe() : graph<Tvalue>()
{

}

		 	 	 	 	// construit le graphe avec un nom et une fonction d'affichage
template <typename Tvalue>
weighted_graphe<Tvalue>::
weighted_graphe(
				string aName,
				function<void (Tvalue)> display_vertice_value_function
	 	 	 ) :
			 graph<Tvalue>(aName, display_vertice_value_function)
{
}
				 // construit le graphe a partir d'une liste de noeuds et
				 // d'adjacences (avec predecesseurs et successeurs).
				 // ATTENTION : les objets internes sont des copies des paramètres

template <typename Tvalue>
weighted_graphe<Tvalue>::
weighted_graphe(
				string aName,
				Tvertices_map<Tvalue>& vertices,
				Tadjacences_map& adjacences,
				Tweighted_edges_map& edges_map,
				function<void (Tvalue)> display_vertice_value_function
			   ) : graph<Tvalue>(aName, vertices, adjacences, display_vertice_value_function)
{
	this->weighted_edges_map = edges_map ; // copie
}

// construit le graphe a partir d'une liste de noeuds et
// d'adjacences (avec predecesseurs et successeurs).
// ATTENTION : les objets internes sont des copies des paramètres
template <typename Tvalue>
weighted_graphe<Tvalue>::
weighted_graphe(
				string aName,
				Tvertices_map<Tvalue>& vertices,
				Tadjacences_map& adjacences,
				Tweighted_edges_map& edges_map,
				function<void (Tvalue)> display_vertice_value_function,
				function<string (Tvalue)> toString_value_function
				) : graph<Tvalue>(aName, vertices, adjacences,
						          display_vertice_value_function,
								  toString_value_function)
{
	this->weighted_edges_map = edges_map ; // copie
}


template <typename Tvalue>
weighted_graphe<Tvalue>::
weighted_graphe(
				string aName,
				Tvertices_map<Tvalue>& vertices,
				function<void (Tvalue)> display_vertice_value_function
	 	 	 ) : graph<Tvalue>(aName, vertices, display_vertice_value_function)
{
}

				// constructeur par copie
template <typename Tvalue>
weighted_graphe<Tvalue>::
weighted_graphe(const weighted_graphe<Tvalue>& g) : graph<Tvalue>(g)
{
	this->weighted_edges_map = g.weighted_edges_map ;
}

				// DESTRUCTEUR
template <typename Tvalue>
weighted_graphe<Tvalue>::
~weighted_graphe()
{
}

template <typename Tvalue>
graph<Tvalue>* weighted_graphe<Tvalue>::
copy_graph()
{
	return new weighted_graphe<Tvalue>(*this) ;
}

	// accesseurs

template <typename Tvalue>
Tweighted_edges_map& weighted_graphe<Tvalue>::
get_edges_map()
{
	return this->weighted_edges_map ;
}

template <typename Tvalue>
Tweighted_edges_map weighted_graphe<Tvalue>::
get_copy_edges_map()
{
	return this->weighted_edges_map ;
}

		// mutateurs
template <typename Tvalue>
void weighted_graphe<Tvalue>::
set_edges_map(Tweighted_edges_map& edges_map)
{
	this->weighted_edges_map = edges_map ;
}


template <typename Tvalue>
float weighted_graphe<Tvalue>::
degree(Tvertice_identifier ident)
{
	if (!this->in_graph(ident))
		return -1 ;

	Tvertice_identifiers_set vertices ;

							// recupere les noeuds adjacents a ident
	this->adjacent_vertices(ident, vertices) ;

							// somme les poids des aretes vers les noeuds
							// adjacents a ident
	float degre = 0 ;

	for (Tvertice_identifiers_set_iterator it = vertices.begin() ;
		 it != vertices.end() ; it++)
	{
		pair<Tvertice_identifier, Tvertice_identifier> p = make_pair(ident, *it) ;
		degre += this->get_edges_map().at(p) ;
	}

	// cout << "degre = " << degre << endl ;
	return degre ;
}

// sauvegarde le graph en format dot

template <typename Tvalue>
void weighted_graphe<Tvalue>::
save_graph_dot(string name, bool vertice_ident, bool vertice_label)
{
	this->save_graph_dot2(name, vertice_ident, vertice_label, false) ;
}

template <typename Tvalue>
void weighted_graphe<Tvalue>::
save_graph_dot2(string graph_name,
				bool vertice_ident,
				bool vertice_label,
				bool directed_graph)
{
	string file_name = "./" + graph_name + ".dot" ;
	ofstream fichier(file_name, ios::out | ios::trunc);

	string kind_of_graph = "Graph";
	string kind_of_edge  = " -- ";

	if (directed_graph)
	{
		kind_of_graph = "diGraph" ;
		kind_of_edge  = " -> " ;
	}

	if(fichier)
	{
		fichier << kind_of_graph  + " " + graph_name << "{" << endl ;
					 // sauve les noeuds (vertices)

											// titre du graph
		fichier << "labelloc=top;" << endl ;
		// fichier << "labeljust=left;" << endl ;
		fichier << "fontsize=20;" << endl ;
		fichier << "label=" << "\"" << graph_name << "\";" << endl ;

					 // pour chaque noeud x
	for (Tvertice_identifiers_set_iterator x = this->get_vertice_identifiers_list().begin() ;
			x != this->get_vertice_identifiers_list().end() ; x++)
	{
		string nom_noeud = graph<Tvalue>::toString_value(this->get_vertice_value(*x)) ;
		string identifiant_noeud = to_string(*x) ;
		string label_noeud = this->get_vertice_label(*x) ;

		string label_node = nom_noeud ;

		if (vertice_ident)
		{
			label_node = identifiant_noeud + ":" + label_node ;
		}

		if (vertice_label)
		{
			label_node = label_node + ":" + label_noeud ;
		}

		string option_label = identifiant_noeud + " [label= \"" + label_node + "\"" + "]" ;

		fichier << option_label << ";" << endl ;
	}

					// sauve les aretes (ici arcs)

						// pour chaque noeud x
	for (Tvertice_identifiers_set_iterator x = this->get_vertice_identifiers_list().begin() ;
			x != this->get_vertice_identifiers_list().end() ; x++)
	{
		Tvertice_identifiers_set& successors_x = this->get_successors(*x) ;
		Tvertice_identifiers_set_iterator y    = successors_x.begin() ;

							// pour chaque successeur y de x
		for ( ; y != successors_x.end() ; y++)
		{
			float weight = this->get_edges_map()[make_pair(*x,*y)] ;

			if (vertice_ident)
				fichier << to_string(*x) << kind_of_edge << to_string(*y) ;
			else
				fichier << graph<Tvalue>::toString_value(this->get_vertice_value(*x))
						<< graph<Tvalue>::toString_value(this->get_vertice_value(*y)) ;

			fichier << 	" [label = " << weight << "] ;" << endl ;

		}
	}

	fichier << "}" << endl ;
	fichier.close();
	}
	else
		cerr << "Erreur a l'ouverture !" << endl;
}

#endif /* WEIGHTED_GRAPHE_HPP_ */





