/**
 *  Version 10.0 (includes weighted graphs)
 *
 * file : MINERLC.CPP
 * from a program made by Benjamin Negrevergne <benjamin@neb.dyn.cs.kuleuven.be>
 * @brief ParaMiner instance for mining closed frequent itemsets (fim)
 *
 *
 * MINER LOCAL CLOSURES
 * Henry SOLDANO (LIPN University Paris 13)
 * Guillaume SANTINI (LIPN University Paris 13)
 * Dominique BOUTHINON(LIPN, University Paris 13)
 *
 * THIS PROGRAM MUST BE COMPILED WITH VERSION C++11
 * WITH OPTIONS THAT ALLOW FOR <functionial>
 *
 *
 */
#include <cstdlib>
#include <algorithm>
#include <vector>
#include <sstream>
#include <fstream>
#include <iostream>
#include <sys/time.h>
#include <time.h>
#include <ctime>
#include <stdio.h>
#include <unistd.h>
#include <set>
#include <string>

#include "statsModularity.hpp"
#include "graph.hpp"
#include "weighted_graphe.hpp"
#include "directed_graph.hpp"
#include "types_graph.hpp"
#include "abstract_local_itemsets.hpp"
#include "pattern.hpp"
#include "database.hpp"
#include "read_write_data2.hpp"
#include "display_data.hpp"
#include "parameters.hpp"
#include "miner_local.hpp"
#include "utils.hpp"
#include "utils_divers.hpp"
#include "common_functions.hpp"

using std::cout;
using std::cerr;
using std::endl;

#define PERMUTE_ITEMS

int ELEMENT_RANGE_END;

set_t permutations;

using namespace std;


	// retourne la description d'un noeud

void get_description_noeud(
							 const TransactionTable& tt,		// table transactions
							 const vector<string>& items_names,	// nom des attributs
							 Tvertice_identifier noeud,			// ident noeud
							 string& description				// description noeud
							)
{
	description = "" ;

	Transaction t = tt[noeud] ;

		 	 	 	 	 	 // pour chaque item  de la Transaction
	for(Transaction::const_iterator attribut = t.begin(); attribut != t.end(); ++attribut)
	{
		// cout << "n° attribut interne = " << *attribut << endl ;
		// cout << "n° attribut         = " << permutations[*attribut] << endl ;
		// cout << "nom attribut        = " << items_names[permutations[*attribut]] << endl ;

		if (attribut == t.begin())
			description = items_names[permutations[*attribut]] ;
		else
							// t_it designe un item (index) de la transaction
		description =  description + "," + items_names[permutations[*attribut]] ;
	}
}


void display_descriptions_noeuds(
								 const TransactionTable& tt,		// table transactions
								 const vector<string>& items_names	// nom des attributs
								 )
{

	cout << "debut description" << endl ;
	cout << "tt.size() " << tt.size() << endl ;

	string d = "" ;

	for (int i = 0 ; i < tt.size() ; i++)
	{
		get_description_noeud(tt, items_names, i, d) ;
		cout << "noeud " << i << " : " << d << endl ;
	}
}



void build_descriptions_noeuds(
								 const TransactionTable& tt,		// table transactions
								 const vector<string>& items_names,	// nom des attributs
								 vector<string>& descriptions_noeuds
							  )
{

	for (int i = 0 ; i < tt.size() ; i++)
	{
		string d = "" ;
		get_description_noeud(tt, items_names, i, d) ;
		descriptions_noeuds.push_back(d) ;
	}
}



/**
 *
 * (appelé depuis paraminer (et non pas paraminer_local_itemsets))
 */

void element_print(const element_t element)
{
  cout<< permutations[element] ;
}

/*
void original_element_string(string& s, const element_t element,
				   	         const vector<string>& items_names)
{

  if (element >= permutations.size())
  {
	  s << "problem" << endl ;
  }

  if (permutations[element] >= items_names.size())
  {
	  s << "problem 2" << endl ;
  }

  s << items_names[permutations[element]] ;
}
*/


void get_original_element(
						  const element_t element,
	        			  const vector<string>& items_names,
						  string& name
						 )
{
	if (element >= permutations.size())
	{
		  cerr << "problem" << endl ;
	}

	if (permutations[element] >= items_names.size())
	{
		  cerr << "problem 2" << endl ;
	}

	name = items_names[permutations[element]] ;
}


void original_element_print(const element_t element,
				   	        const vector<string>& items_names)
{

  if (element >= permutations.size())
  {
	  cout << "problem" << endl ;
  }

  if (permutations[element] >= items_names.size())
  {
	  cout << "problem 2" << endl ;
  }

  cout<< items_names[permutations[element]] ;
}

void original_element_print_cerr(const element_t element,
				   	             const vector<string>& items_names)
{

  if (element >= permutations.size())
  {
	  cerr << "problem" << endl ;
  }

  if (permutations[element] >= items_names.size())
  {
	  cerr << "problem 2" << endl ;
  }

  cerr << items_names[permutations[element]] ;
}

typedef struct
{
  bool operator()(const pair<int,int> &a, const pair<int,int> &b)
  {
    return a.second < b.second;
  }
}
pair_gt_second_t;


// ordonne les items en fonction de leur frequence

vector<pair<int,int> > sort_items(const vector<vector<int> >& item_occurrences)
{
	vector<pair<int,int> > frequences ;

									// pour chaque item
	for (unsigned int item = 0 ; item < item_occurrences.size() ; item++)
	{
		int frequence = item_occurrences[item].size() ;
		frequences.push_back(pair<int,int>(item, frequence)) ;
	}
	// end for in item

	return frequences ;
}
// end compute_frequences_initial_objects



/*   calcule la frequence des items
 * 	 ne retient que les items dont la frequence est superieure ou
 * 	 egale a minsup
 * 	 ordonne les items par frequence croissante
 * 	 permutation[item] est un indice attribué à l'item conservé
 */
int compute_permutation_by_frequency(
										vector<vector<int> >& transacts,
										set_t* permutations,
										int minsup,
										vector<vector<int>>& initial_objects
									 )
{

	vector<vector<int> > item_occurrences ;
								// item_occurrences[item] contient les numéros
								// des transactions contenant item

	element_t max_element = -1;

								// Pour chaque transaction
	for (unsigned int i = 0 ; i < transacts.size() ; i++)
	{
		vector<int>& transact = transacts.at(i) ;

								// pour chaque item de la transaction
		for (unsigned int j = 0 ; j < transact.size(); j++)
		{
			int& item = transact.at(j) ;

			if(item > max_element)		// 1ere rencontre de l'item
			{
				  max_element=item;
				  item_occurrences.resize(max_element+1);
			}
			  	  	  	  	  	  	  	  //  item apparait dans transaction i
			item_occurrences[item].push_back(i);
		 }
	}

	vector<pair<int,int> > freq_table = sort_items(item_occurrences) ;

	 pair_gt_second_t gt;
	 sort(freq_table.begin(), freq_table.end(),gt);

						  // freq_table est trie en ordre croissant selon
						  // la frequence des items

	permutations->resize(max_element+1);

						 // on attribue a chaque item un index dans
						 // permutation
						 // en ne conservant que les items suffisamment
						 // representés
	int item_index = 0;

	for(unsigned int i = 0; i < freq_table.size(); i++)
	{
		if(freq_table[i].second >= minsup)
		  (*permutations)[freq_table[i].first] = item_index++;
		else
		  (*permutations)[freq_table[i].first] = -1;
	}


	#ifdef PERMUTE_ITEMS

	#else
				  /* reset permutations (dirty) */
	for(int i = 0; i < permutations->size(); i++)
	{
		(*permutations)[i] = i;
	}
	#endif

	return item_index;
}


	// teste si |ext(pattern + augmentation))| >= min_sup
	// (pattern or clo(pattern))
int membership_oracle_on_O2(
							const Toccurences &pattern_occurences,
							const Toccurences &extension_occurences,
							int min_sup
			  	  	       )
{
	Toccurences occs ;

	setIntersection(&occs, pattern_occurences, extension_occurences) ;

	int support = occs.size() ;

	if (support >= min_sup)
		return support ;

	return 0;
}



set_t clo(const set_t &set, 					// pattern à clore
		  const closure_data_t &data)
{
  return support_based_closure(set, data.set_support, data.support);
}


		// construit la table des transactions (tt) à partir des
		// transactions originales (transacts) et des permutations
		// effectuees
element_t build_transaction_table(TransactionTable *tt,
							     const vector<vector<int> >& transacts,
								 const set_t& permutations)
{
	element_t max_element = 0;

	int nb_items = 0;
	int nb_trans = 0;

	  	  	  	  	  	  	  	  // pour chaque transaction de transacts
	for (unsigned int i = 0 ; i < transacts.size() ; i++)
	{
		Transaction t;
								// pour chaque item de transact
		for (unsigned int j = 0 ; j < transacts.at(i).size() ; j++)
		{

			  int item = transacts.at(i).at(j) ;

			  element_t p = permutations[item];   // p est le numéro d'item interne

			  if(p != -1)
			  {
				  t.push_back(p);
			  }

			  if(p > max_element)
				  max_element=p;

		    nb_items++ ;
		}

			t.limit=t.size();

			/*
			#ifdef TRACK_TIDS
			  if(show_tids)
				  t.tids.push_back(nb_trans);
			#endif
			*/

		   nb_trans++;
		   t.weight = 1;

		   #ifdef SORT_DATABASE
				  sort(t.begin(), t.end());
		   #else

				   #ifndef NDEBUG
					  assert(is_sorted(t));
				   #endif

		   #endif //SORT_DATABASE

		   tt->push_back(t);

	  }
	  // fin pour chaque transaction

	  cout  << "# loaded items                     = " << nb_items << endl ; 			
	  cout  << "# loaded transactions              = " << nb_trans << endl ;

	  tt->max_element = max_element;

	  return max_element;
}


void write_final_messages(statsModularity& statsMod,
		                  int nb_closed_patterns, int nb_rejected_closed_patterns,
						  int num_pattern, time_t& debut, time_t & fin)
{
	 cout << "# ===================================================================" << endl ;
	 cout << "# closed pattern(s)                                              = " << nb_closed_patterns          << endl ;
	 cout << "# rejected closed pattern(s)                                     = " << nb_rejected_closed_patterns << endl ;
	 cout << "# septuple(s) (c, ext(c), ext1(c), ext(l), ext1(l), ext2(l), l)  = " << num_pattern  << endl ;
	 cout << "# nec                                                            = " << statsMod.getNec() << endl ;
	 cout << "# f                                                              = " << statsMod.getF()   << endl ;
	 cout << "# lme                                                            = " << statsMod.getNbLme() << endl ;
	 cout << "# lm                                                             = " << statsMod.getNbLm() << endl ;
	 cout << "# begin                                                          = " << ctime(&debut) ;
	 cout << "# end                                                            = " << ctime(&fin)  ;
	 cout << "# duration                                                       = " << difftime(fin,debut) << " seconds." << endl ;
}


void write_minerLC_parameters(string program, string version,
				  string name_data_file, bool save_graphs, bool modularity_on, 
				  float modularity,
				  bool local, int min_sup,
				  bool build_triangles_graph, bool build_edges_graph,
				  abstraction& a1, abstraction& a2,
				  bool a2_active, bool communauty_abstraction,
				  bool supports_demandes,
				  int k_clique, int k_crown, bool directed_graph, bool weighted_graph)
{
  cout << "# program                          = " << program		  	   << endl ;
  cout << "# version                          = " << version		  	   << endl ;
  cout << "# data file                        = " << name_data_file 	   << endl ;
  cout << "# sauvegarde graphes               = " << save_graphs 	  	   << endl ;
  cout << "# local modularity                 = " << modularity_on		   << endl ;
  cout << "# local modularity  threshold      = " << modularity			   << endl ;
  cout << "# local                            = " << local       	  	   << endl ;
  cout << "# min_sup                          = " << min_sup       	   	   << endl ;
  cout << "# build_triangles_graph            = " << build_triangles_graph << endl  ;
  cout << "# build_edges_graph                = " << build_edges_graph     << endl ;
  write_abstractions(a1, "(a1)") ;
  write_abstractions(a2, "(a2)") ;
  cout << "# a2_active                        = " << a2_active 				<< endl ;
  cout << "# communauty_abstraction (ca)      = " << communauty_abstraction << endl ;
  cout << "# supports demandes                = " << supports_demandes      << endl ;
  cout << "# k-clique                         = " << k_clique           	<< endl ;
  cout << "# k-crown                          = " << k_crown                << endl ;
  cout << "# directed_graph                   = " << directed_graph	    	<< endl ;
  cout << "# weighted_graph                   = " << weighted_graph	    	<< endl ;
}


			//	gestion des options du programme
bool handle_minerLC_parameters(
					   int argc,
					   char* argv[],
					   bool& save_graphs,
					   bool& modularity_on,
					   float& modularity,			// modularite locale
					   bool& local,
					   int& min_sup,
					   bool& build_triangles_graph,
					   bool& build_edges_graph,
					   abstraction& a1,
					   abstraction& a2,
					   bool& a2_active,			// temoigne que a2 differe de a1 et doit etre applique
					   bool& communauty_abstraction,
					   bool& support_demande,
					   int&  k_clique,
					   int&  k_crown,
					   bool& directed_graph,
					   bool& weighted_graph
					  )
{
	bool ok = true ;

	int i = 2 ;							// 2 premiers parametres vus avant


	while (i < argc && ok)
	{
		// cout << "argument " << i << " = " << argv[i] << endl ;

		if(string(argv[i]) == "-save_graphs")
		{
			save_graphs = true ;
			i++ ;
		}
		else if(string(argv[i]) == "-lm")
		{
			modularity = atof(argv[i+1]) ;
			modularity_on = true ;
			i = i + 2 ;
		}
		else if(string(argv[i]) == "-min_sup")
		{
			min_sup = atoi(argv[i+1]) ;
			i = i + 2 ;
		}
		else if(string(argv[i]) == "-tg")
		{
			build_triangles_graph = true ;
			local = true ;
			i++ ;
		}
		else if(string(argv[i]) == "-eg")
		{
			build_edges_graph   = true ;
			local = true ;
			i++ ;
		}
		else if(string(argv[i]) == "-local")
		{
			local = true ;
			i++ ;
		}
		else if(string(argv[i]) == "-clique")			// k communautes de cliques
		{
			k_clique = atoi(argv[i+1]) ;
			k_crown = 0 ;								// k_crown pas possible si k-clique

			if (k_clique < 3)
			{
				cout << "!!! ERROR !!! " << endl ;
				cout << "-clique " << k_clique << endl ;
				cout << "the argument must be greater than 2" << endl ;
				ok = false ;
			}

			local = true ;
			i = i + 2 ;
		}
		else if(string(argv[i]) == "-crown")			// k communautes de couronnes
		{
			k_crown = atoi(argv[i+1]) ;

			k_clique = 0 ;

			if (k_crown < 3)
			{
				cout << "!!! ERROR !!! " << endl ;
				cout << "-crown " << k_crown << endl ;
				cout << "the argument must be greater than 2" << endl ;
				ok = false ;
			}

			local = true ;
			i = i + 2 ;
		}
		else if(string(argv[i]) == "-ca")
		{
			communauty_abstraction = true ;
			i++ ;
		}
		else if(string(argv[i]) == "-directed")
		{
			directed_graph = true ;
			i++ ;
		}
		else if(string(argv[i]) == "-weighted")
		{
			weighted_graph = true ;
			i++ ;
		}
		else if(string(argv[i]) == "-s")			// supports au lieu d'extensions
		{
			support_demande = true ;
			i++ ;
		}
		else if (handle_abstractions(argv, a1, i, "1")) 
		{
		}
		else if (handle_abstractions(argv, a2, i, "2")) 
		{
		}
		else
		{
			ok = false ;
			cerr << "parameter " << argv[i] << " not recognized" << endl ;
		}
	}
	 						// NOTE : k_dense == 3 <=> abstraction triangles

							// l'abstraction a2 doit aussi contenir a1,
							// sauf si a2 ne contient qu'une abstraction en
							// composantes connexes (l'abstraction a1 sera
							// toujour satisfaite dans a2).

							// (pour chaque abstraction (degre, cc, etc...)
							// il faut retenir pour a2 la plus contraignante entre a1 et a2

	if (ok)
		return set_a2_abstraction(a1, a2, directed_graph, a2_active) ;

	return false ;
}


// ============================== MAIN ========================================

int main(int argc, char* argv[])
{
  string version = "10.0" ;
  vector<string> objects_names ;
  vector<string> items_names ;
  vector<vector<int> > transacts ;
  vector<vector<int> > initial_objects ;
  Tadjacences_map  adjacences ;
  vector<string> description_noeuds ;
  Tweighted_edges_map edges_map ;

  set_t empty_set;
  int nb_closed_patterns          = 0 ;
  int nb_rejected_closed_patterns = 0 ;
  	  	   	  	  	   	  	  	  			// parametres par defaut
  bool local	  				= false ;	// local ou global
  bool communauty_abstraction 	= false ;	// abstraction dans les communautés
  int min_sup	  				= 1     ;	// frequence minimal des patterns a considérer

  abstraction a1 ;
  abstraction a2 ;
  string name_data_file       	= argv[1]   ;
  bool build_triangles_graph  	= false ;
  bool build_edges_graph		= false ;
  bool a2_active             	= false ;
  bool support_demande          = false ;
  int  k_clique                 = 0     ;
  int  k_crown                  = 0     ;
  bool graphe_oriente			= false ;
  bool graphe_pondere		    = false ;
  bool save_graphs              = false ;
  bool label_noeuds 			= false ;
  float modularity				= 0 ;
  bool modularity_on			= false ;
  statsModularity statsMod ;

  if (!handle_minerLC_parameters(argc, argv, save_graphs, modularity_on, modularity, local,
		  	  	  	  	  	  	  min_sup, build_triangles_graph,
								  build_edges_graph, a1, a2, a2_active,
								  communauty_abstraction,
								  support_demande, k_clique, k_crown, graphe_oriente, graphe_pondere))
	  return 0 ;

  label_noeuds = save_graphs ;

  write_minerLC_parameters(argv[0], version, name_data_file, save_graphs, modularity_on, modularity,
		  	  	  	  	  local, min_sup,
		  	  	   	      build_triangles_graph, build_edges_graph, a1, a2,
						  a2_active, communauty_abstraction,
						  support_demande, k_clique, k_crown, graphe_oriente, graphe_pondere) ;


  // cerr << endl << "reading data..." << endl ;

  read_data_new_ri(graphe_oriente, graphe_pondere, name_data_file, objects_names,
		  	  	   items_names, transacts, adjacences, edges_map) ;

  // afficher_adjacences(adjacences) ;

  // cerr << "apres creation" << endl ;
  // afficher_adjacences(adjacences) ;

  // displayVectorVector("transacts", transacts) ;
  // afficher_adjacences(adjacences) ;


  // cerr << "computing permutations..." << endl ;

  compute_permutation_by_frequency(transacts,
		  	  	  	  	  		   &permutations,
								   min_sup,
								   initial_objects);
  // g->visu_graph_dot("graphe_initial", true) ;

  // displayVector("objects_names", objects_names) ;
  // cout << endl ;

  // displayVector("items_names", items_names) ;
  // cout << endl ;

  element_t max = build_transaction_table(&tt, transacts, permutations);

  // cerr << "reverse permutations " << endl ;

  	  	  	  	  	  	  // inverse indices et valeurs dans permutations
 reverse_permutations(&permutations);

 // display_descriptions_noeuds(tt, items_names) ;

  if (label_noeuds)
	  build_descriptions_noeuds(tt, items_names, description_noeuds) ;


  // cerr << "building initial graph..." << endl ;

  graph<string>*initial_graph = create_initial_graph(name_data_file, label_noeuds,
		  	  	  	  	  	  	  	  	  	  	  	 objects_names,
  		  	  	  	  	  	  	  	      	  	     description_noeuds,
													 adjacences, edges_map,
													 graphe_oriente,
													 graphe_pondere,a1, a2) ;

  // cerr << "visualizing graph..." << endl ;

 // initial_graph->visu_graph_dot("graphe_initiale", true, true) ;

 // cout << "nb edges = " << initial_graph->nb_edges() << endl ;

 ELEMENT_RANGE_END = max+1;

  // cerr << "tranposing..." << endl ;

  transpose(tt, &ot);

  time_t debut ;
  time(&debut) ;


  try
  {
	  replace(name_data_file.begin(), name_data_file.end(), '.', '_') ;

	  cerr << "searching patterns..." << endl ;

  	  int num_pattern = abstract_local_itemsets(name_data_file,
  			  	  	  	  	  	  	  	  	    save_graphs,
												statsMod,
												modularity_on,
												modularity,
												support_demande,
  			  	  	  	  	  	  	  	  	  	a1, a2, a2_active,
  			  	  	  	  	  	  	    		communauty_abstraction,
												min_sup, local,
												build_triangles_graph,
												build_edges_graph,
												k_clique,
												k_crown,
	   	  	  	  	  	    				    initial_graph, graphe_oriente,
												objects_names, items_names,
												permutations,
											    nb_closed_patterns,
										        nb_rejected_closed_patterns);
  	  time_t fin ;
  	  time(&fin) ;

  	  write_final_messages(statsMod, nb_closed_patterns, nb_rejected_closed_patterns,
  					   	   num_pattern, debut, fin) ;
  	  cerr << "end." << endl ;
  }
  catch(exception& e)
  {
	  cerr << e.what() << endl ;
  }
  catch(string& s)
  {
	  cerr << s << endl ;
  }

  // cerr << "end of the program " << argv[0] << endl ;

  return 0 ;
}
