/*
 *	MINERLC10.0
 * read_write_data2.cpp
 *
 *  Created on: 18 mars 2015
 *  Author: Dominique Bouthinon
 *
 */

#include <cstdlib>
#include <algorithm>
#include <vector>
#include <sstream>
#include <fstream>
#include <iostream>
#include <iterator>
#include <set>
#include <string>
#include <tuple>

#include "read_write_data2.hpp"
#include "types_graph.hpp"
#include "utils_divers.hpp"

using namespace std;

/**
 *  get the int values between the current line and the line containing stop
 */
void get_transactions(
				      ifstream& input,
					  int nb_objects,
				      vector<vector<int> >& values
			         )
{
	cerr << "reading transactions..." << endl ;

								// pour chaque objet (ligne)
	for (int i = 0 ; i < nb_objects ; i++)
	{
		 vector<int>    num_items ;
		 vector<string> lines ;
		 string line ;

		 getline(input, line) ;
		 // cout << line << endl ;

		 split(lines, line, " ")   ;

		 if (lines.size() > 1)
		 {
			 get_ints(lines[1], ',', num_items) ;
		 }
		 values.push_back(num_items) ;
	}
}


// creer les predecesseurs a partir de la liste des successeurs

void creer_predecesseurs(Tadjacences_map& adjacences)
{
							// pour chaque noeud n
	for(Tadjacences_map_iterator n = adjacences.begin(); n != adjacences.end(); n++)
	{
		// cerr << "noeud : " << n->first << endl ;

							// pour chaque successeur s du noeud n
		for (Tvertice_identifiers_set_iterator s = n->second.successors.begin() ;
			 s != n->second.successors.end() ; s++)
		{
			// cerr << "successeur : " << *s << endl ;

							// noeud n (en fait n->first
							// integre les predeccesseurs de s
			adjacences[*s].predeccessors.insert(n->first) ;
		}
	}
}


void get_adjacences_orientees_ponderees(
					 	 	 	 	 	 ifstream& input,
										 int nb_objects,
										 Tadjacences_map& adjacences,
										 Tweighted_edges_map& edges_map
			        		 	 	 	 )
{
	cerr << "reading adjacences..." << endl ;

	for (int i = 0 ; i < nb_objects ; i++)
	{
		 string line ;

		 	 	 	 	 	 // lit une ligne d'adjacences
		 getline(input, line) ;
		 // cout << "line = " << line << "xx" << endl ;

		 vector<string> lines ;
		 Tvertice_identifiers_set adjacent_objects ;
		 set<string> strings ;

		 split(lines, line, " ") ;

		 Tvertice_identifier ident = stoi(lines[0]) ;	// ident est le noeud concern�
		 // cout << ident << endl ;

		 	 	 	 	 	 	 	 	 	 // lit les noeuds adjacents et le poids
		 	 	 	 	 	 	 	 	 	 // des aretes
		 if (lines.size() > 1)
		 {
			 get_strings(lines[1], ',', strings) ;
		 }

											 // adjacent_objects contients
											 // les noeuds adjacents � ident
											 // chaque noeud de la forme
											 // 3:1.5   (noeud 3, poids 1.5)

		 // displaySet("strings", strings) ;
		 									 // distribue les identifiants des noeuds
		 									 // adjacents en predecesseurs et successeurs
		 	 	 	 	 	 	 	 	 	 // et stoke les poids des aretes
		 Tadjacences adja ;

		 set<string>::iterator it     = strings.begin() ;
		 set<string>::iterator it_end = strings.end() ;

		 // cout << "ident = " << ident << endl ;

		 for ( ; it != it_end ; it++)
		 {
			string s = *it ;
			// cout << "s = " << s << endl ;

			vector<string> ss ;

			split(ss, s, ":") ;

			Tvertice_identifier v = stoi(ss[0]) ;

			float weight = stof(ss[1]) ;

			edges_map.insert(make_pair(make_pair(ident, v), weight)) ;

			// cout << ident << "-" << v << " : " << weight << endl ;

			adja.successors.insert(v) ;
		 }
		 adjacences.insert(make_pair(ident, adja)) ;
	}
}

void get_adjacences_ponderees(
					 	 	 	 ifstream& input,
								 int nb_objects,
								 Tadjacences_map& adjacences,
							 	 Tweighted_edges_map& edges_map
			        		 )
{
	cerr << "reading adjacences..." << endl ;

	for (int i = 0 ; i < nb_objects ; i++)
	{
		 string line ;

		 	 	 	 	 	 // lit une ligne d'adjacences
		 getline(input, line) ;
		 // cout << "line = " << line << "xx" << endl ;

		 vector<string> lines ;
		 Tvertice_identifiers_set adjacent_objects ;
		 set<string> strings ;

		 split(lines, line, " ") ;

		 Tvertice_identifier ident = stoi(lines[0]) ;	// ident est le noeud concern�
		 // cout << ident << endl ;

		 	 	 	 	 	 	 	 	 	 // lit les noeuds adjacents et le poids
		 	 	 	 	 	 	 	 	 	 // des aretes
		 if (lines.size() > 1)
		 {
			 get_strings(lines[1], ',', strings) ;
		 }

											 // adjacent_objects contients
											 // les noeuds adjacents � ident
											 // chaque noeud de la forme
											 // 3:1.5   (noeud 3, poids 1.5)

		 // displaySet("strings", strings) ;
		 									 // distribue les identifiants des noeuds
		 									 // adjacents en predecesseurs et successeurs
		 	 	 	 	 	 	 	 	 	 // et stoke les poids des aretes
		 Tadjacences adja ;

		 set<string>::iterator it     = strings.begin() ;
		 set<string>::iterator it_end = strings.end() ;

		 // cout << "ident = " << ident << endl ;

		 for ( ; it != it_end ; it++)
		 {
			string s = *it ;
			// cout << "s = " << s << endl ;

			vector<string> ss ;

			split(ss, s, ":") ;

			Tvertice_identifier v = stoi(ss[0]) ;

			float weight = stof(ss[1]) ;

			edges_map.insert(make_pair(make_pair(ident, v), weight)) ;

			// cout << ident << "-" << v << " : " << weight << endl ;

			if(v < ident)
				adja.predeccessors.insert(v) ;
			else
				adja.successors.insert(v) ;
		 }
		 adjacences.insert(make_pair(ident, adja)) ;
	}
}



/**
 *  get the adjacences of each node
 *  NOUVEAU TRAITEMENT VALABLE POUR GRAPHE ORIENTE OU NON
 */
void get_adjacences_orientees(
					 	 	 	 ifstream& input,
							 	 int nb_objects,
								 Tadjacences_map& adjacences
			        		 )
{
	cerr << "reading adjacences... " << endl ;

	for (int i = 0 ; i < nb_objects ; i++)
	{
		 string line ;

		 getline(input, line) ;
		 // cout << "line = " << line << "xx" << endl ;

		 vector<string> lines ;
		 Tvertice_identifiers_set adjacent_objects ;

		 split(lines, line, " ") ;
		 	 	 	 	 	 	 	 	 	 	 	 	 // recupere nom noeud (ident)
		 Tvertice_identifier ident = stoi(lines[0]) ;
		 // cout << ident << endl ;

		 	 	 	 	 	 	 	 	 	 	 	 	 // recupere noeuds adjacents a ident
		 if (lines.size() > 1)
		 {
			 get_ints(lines[1], ',', adjacent_objects) ;
		 }

		 // displaySet("adja", adjacent_objects) ;

		 									 // distribue noeuds adjacents
		 									 // comme successeurs
		 Tadjacences adja ;

		 Tvertice_identifiers_set_iterator it     = adjacent_objects.begin() ;
		 Tvertice_identifiers_set_iterator it_end = adjacent_objects.end() ;

		 // cout << "ident = " << ident << endl ;

		 for ( ; it != it_end ; it++)
		 {
				// cout << "*it = " << *it << endl ;
				adja.successors.insert(*it) ;
		 }

		 adjacences.insert(make_pair(ident, adja)) ;
	}
}



/**
 *  get the adjacences described between the current line and the line containing stop
 */
void get_adjacences(
					 ifstream& input,
					 int nb_objects,
					 Tadjacences_map& adjacences
			        )
{
	cerr << "reading adjacences..." << endl ;

	for (int i = 0 ; i < nb_objects ; i++)
	{
		 string line ;

		 getline(input, line) ;
		 // cout << "line = " << line << "xx" << endl ;

		 vector<string> lines ;
		 Tvertice_identifiers_set adjacent_objects ;

		 split(lines, line, " ") ;

		 	 	 	 	 	 	 	 	 	 	 	 	// recupere numero noeud (ident)
		 Tvertice_identifier ident = stoi(lines[0]) ;
		 // cout << ident << endl ;

		 	 	 	 	 	 	 	 	 	 	 	 	// recupere noeuds adjacents a ident
		 if (lines.size() > 1)
		 {
			 get_ints(lines[1], ',', adjacent_objects) ;
		 }

		 // displaySet("adja", adjacent_objects) ;

		 									 // distribue les noeuds adjacents
		 									 // en predecesseurs et successeurs
		 Tadjacences adja ;

		 Tvertice_identifiers_set_iterator it     = adjacent_objects.begin() ;
		 Tvertice_identifiers_set_iterator it_end = adjacent_objects.end() ;

		 // cout << "ident = " << ident << endl ;

		 for ( ; it != it_end ; it++)
		 {
				// cout << "*it = " << *it << endl ;

			if(*it < ident)
				adja.predeccessors.insert(*it) ;
			else
				adja.successors.insert(*it) ;
		 }

		 adjacences.insert(make_pair(ident, adja)) ;
	}
}

	// lit les donnees pour minerLC10 depuis un fichier
void read_data_new_ri(
						bool graphe_oriente,
						bool graphe_pondere,
						const string fileName,
						vector<string>& object_names,
						vector<string>& item_names,
						vector<vector<int> >& transacts,
						Tadjacences_map& adjacences,
						Tweighted_edges_map& edges_map
			         )
{
	    ifstream input(fileName , ifstream::in); 		// ouvre fichier

	    if (input)
	    {
			string line;

			getline(input, line) ;						// lecture commentaire
			// cout << line << endl ;

			getline(input, line) ;						// lecture noms objets
			// cout << "line = " << line << endl ;

			split(object_names, line, " |") ;
			trimStringVector(object_names) ;
														// le dernier objet peut etre vide
														// si la ligne contenant le nom des objets
														// finissait par |
			int nb_objects      = object_names.size() ;
			string& last_object = object_names[nb_objects-1] ;

			if (last_object.size() == 0)
			{
				object_names.pop_back() ;
			}

			nb_objects = object_names.size() ;

			// displayVector("object_names", object_names);
			// cout << endl ;

			getline(input, line) ;						// lecture  attributs
			// cout << line << endl ;

			split(item_names, line, " |") ;
			trimStringVector(item_names)  ;

														// le dernier item peut etre vide
														// si la ligne contenant les items (attributs)
														// finissait par "|"
			int nb_attributes = item_names.size() ;
			string& last_attribute = item_names[nb_attributes-1] ;

			if (last_attribute.size() == 0)
			{
				item_names.pop_back() ;
			}

			// displayVector("item_names", item_names);
			// cout << endl ;
														// lecture transactions
			get_transactions(input, nb_objects, transacts) ;
			// displayVectorVector("transacts", transacts) ;

														// lecture #
			getline(input, line) ;
			// cout << line << endl ;
														// lecture adjacences
			// get_adjacences2(input, nb_objects, adjacences) ;
			// afficher_adjacences(adjacences) ;

			if (graphe_pondere && graphe_oriente)
			{
				get_adjacences_orientees_ponderees(input, nb_objects, adjacences, edges_map) ;
				creer_predecesseurs(adjacences) ;
			}
			else if (graphe_oriente)
			{
				get_adjacences_orientees(input, nb_objects, adjacences) ;
				creer_predecesseurs(adjacences) ;
			}
			else if (graphe_pondere)
				get_adjacences_ponderees(input, nb_objects, adjacences, edges_map) ;
			else
				get_adjacences(input, nb_objects, adjacences) ;

			input.close() ;
	    }
	    else
	    {
	    	cout << "the file " << fileName << " cannot be opened." << endl ;
	    }
}



// lit graphe au nouveau format
void read_graph_data2(
		  	  bool graphe_oriente,
			  const string fileName,
			  vector<string>& object_names,
			  Tadjacences_map& adjacences
		      )
{
	ifstream input(fileName , ifstream::in); 		// ouvre fichier

	if (input)
	{
		string line;

		getline(input, line) ;						// lecture commentaire
		// cout << "line = " << line << endl ;

		getline(input, line) ;						// lecture noms objets
		// cout << "line = " << line << endl ;

		split(object_names, line, " |") ;
		trimStringVector(object_names) ;
													// le dernier objet peut être vide
													// si la ligne contenant le nom des objets
													// finissait par |
		int nb_objects = object_names.size() ;
		string& last_object = object_names[nb_objects-1] ;

		if (last_object.size() == 0)
		{
			object_names.pop_back() ;
		}

		nb_objects = object_names.size() ;

		getline(input, line) ;						// lecture  attributs
		// cout << line << endl ;

												// lecture des transactions

									// pour chaque objet (ligne)
		for (int i = 0 ; i < nb_objects ; i++)
		{

			 string line ;

			 getline(input, line) ;
			 // cout << line << endl ;
		}

											// lecture #
		getline(input, line) ;
		// cout << line << endl ;

		if (graphe_oriente)
		{
			get_adjacences_orientees(input, nb_objects, adjacences) ;
			creer_predecesseurs(adjacences) ;
		}
		else get_adjacences(input, nb_objects, adjacences) ;


		// afficher_adjacences(adjacences) ;
	
	}
	else
	{
		cout << "the file " << fileName << " cannot be opened." << endl ;
	}

	// verifier apres getAdjacences la difference entre objets traites
	// et objets non traites (en utilisant les indices)
	// ajouter successeurs vide et predecesseurs vides aux objets non traites
}






