/**
 *	SUPPORTS_ATTENDUS.CPP 
 *	program modified by Dominique Bouthinon (LIPN, University Paris 13)
 *
 *	Here we assume bi-extensions (e1,e2)
 *
 * THIS PROGRAM MUST BE COMPILED WITH VERSION C++11
 * WITH OPTIONS THAT ALLOW FOR <functionial>
 *
 */  
#include <cstdlib>
#include <algorithm>
#include <vector>
#include <sstream>
#include <fstream>
#include <iostream>
#include <sys/time.h>
#include <time.h>
#include <ctime>
#include <stdio.h>
#include <unistd.h>
#include <set>
#include <string>
#include <tuple>
#include <math.h>

#include "read_supports_input_bi.hpp"
#include "graph.hpp"
#include "directed_graph.hpp"
#include "types_graph.hpp"
#include "read_write_data2.hpp"
#include "parameters.hpp"
#include "common_functions.hpp"
#include "common_supports_functions.hpp"


using std::cout;
using std::cerr;
using std::endl;

using namespace std;


void write_final_messages(int nb_closed_patterns, int nb_rejected_closed_patterns,
						 int num_pattern, time_t& debut, time_t & fin)
{
		
	 cout << "# ===================================================================" << endl ;
	 cout << nb_closed_patterns          << "# closed pattern(s)"              	   << endl ;
	 cout << nb_rejected_closed_patterns << "# rejected closed pattern(s)" 	       << endl ;
	 cout << num_pattern                 << "# triplet(s) (c, e, l)."    		   << endl ;
	 cout << "# begin      = " << ctime(&debut) ;
	 cout << "# end        = " << ctime(&fin)  ;
	 cout << "# duration   = " << difftime(fin,debut) << " seconds." << endl ;
}


void sauvegarder_supports(
					  	  vector<tuple<string, int,int, int> >& supports_observes,
						  double supports_attendus[],
						  double ecarts_types[]
				     	 )
{
   
	int nb_supports = supports_observes.size() ;

	for (int i = 0 ; i < nb_supports ; i++)
	{

		cout  << get<0>(supports_observes[i])<< ", "
			  << get<1>(supports_observes[i])<< ", "
			  << get<2>(supports_observes[i])<< ", "
			  << get<3>(supports_observes[i])<< ", "
			  << supports_attendus[i] << ", "
			  << ecarts_types[i] << endl ;
	}
}

				// compare deux tuples selon son deuxième argument
struct sort_pred
{
    bool operator()(const tuple<string, int, int, int> &left, const tuple<string, int, int, int> &right)
    {
        return get<1>(left) + get<2>(left) > get<1>(right) + get<2>(right);
    }
} ;


// ------------------------------- parameters ----------------------------------

	//	gestion des options du programme
bool handle_supports_attendus_parameters(
										 int argc,
										 char* argv[],
										 abstraction& a,
										 bool& directed_graph,
										 bool& weighted_graph,
										 int& n
		                				)
{

	bool ok = true ;

	int i = 3 ;							// 3 premiers parametres vus avant
	while (i < argc && ok)
	{
		if (handle_abstractions(argv, a, i, "1")) {}
		else if(string(argv[i]) == "-n")
		{
			n = atoi(argv[i+1]) ;
			i = i + 2 ;
		}
		else if(string(argv[i]) == "-directed")
		{
			directed_graph = true ;
			i++ ;
		}
		else if(string(argv[i]) == "-weighted")
			{
				weighted_graph = true ;
				i++ ;
			}
		else
		{
			ok = false ;
			cerr << "parameter " << argv[i] << " not recognized" << endl ;
		}
	}

	return ok ;
}


void write_supports_attendus_parameters(string program,
					               string graph_data_file,
					               string supports_data_file,
					               abstraction& a1,
								   bool directed_graph,
								   bool weighted_graph,
					               int n)
{
  cout << "# program                          = " << program		  	  << endl ;
  cout << "# graph data file                  = " << graph_data_file 	  << endl ;
  cout << "# supports data file               = " << supports_data_file   << endl ;
  cout << "# nombre de tirages                = " << n 					  << endl ;
  cout << "# directed_graph                   = " << directed_graph 	  << endl ;
  cout << "# weighted_graph                   = " << weighted_graph 	  << endl ;
  write_abstractions(a1, "(a1)") ;
  cout << "# " << endl ;
}


		// place les n elements de s dans t
void set_to_tab(set<Tvertice_identifier>& s, Tvertice_identifier t[])
{
	set<Tvertice_identifier>:: iterator it;

	int i = 0 ;
	for( it = s.begin(); it!= s.end(); it++)
	{
	   t[i]= *it;
	   i++ ;
	}
}
// -------------------------------- CALCUL DES SUPPORTS ATTENDUS ---------------

void calcul_supports_attendus_bi(
					   	   	   graph<string>* initial_graph,	// graphe initial
							   bool graphe_dirige,
							   set<Tvertice_identifier> VO, 	// idents noeuds Out
							   set<Tvertice_identifier> VI,	// idents noeuds In
							   vector<tuple<string, int, int, int> >& supports_observes,	// (par ordre croissant)
							   int n,							// nombre d'iterations
							   abstraction a,
							   double* supports_attendus,		// supports abstraits moyens
							   double* ecarts_types             // ecarts types supports abstraits
			    	 	 	 )
{
													// chaque quadruplet de supports_observes contient :
													// element 0 : clos abstrait
													// element 1 : support Out observe
													// element 2 : support In observe
													// element 3 : support abstrait observe


	srand(time(NULL)) ;								// initialisation séquence aleatoire

	int nb_supports = supports_observes.size() ;

													// transforme VO et VI en tableaux
	int nb_out = VO.size() ;
	Tvertice_identifier tVO[nb_out] ;
	set_to_tab(VO, tVO) ;

	int nb_in  = VI.size() ;
	Tvertice_identifier tVI[nb_in] ;
	set_to_tab(VI, tVI) ;

	double* sommes_supports_attendus        = new double[nb_supports] ;
	double* sommes_carres_supports_attendus = new double[nb_supports];

													// initialisation sommets
	for (int j = 0 ; j < nb_supports ; j++)
	{
		  sommes_supports_attendus[j] = 0 ;
		  sommes_carres_supports_attendus[j] =  0 ;
	}

													// pour chaque tirage i
	for (int i = 0 ; i < n ; i++)
	{
		// cout << "i = " << i << endl ;

				// pour chaque motif j
		for (int j = 0 ; j < nb_supports ; j++)
		{

			// cout << "   j = " << j << endl ;
			graph<string>* g1 ;

									// copie du graphe initial
			if (graphe_dirige)
			{
				directed_graph<string>* gp = dynamic_cast<directed_graph<string>* > (initial_graph) ;
				g1 = new directed_graph<string>(*gp) ;
			}
			else
				g1 = new graph<string>(*initial_graph) ;

			// cout << get<0>(supports_observes.at(j)) << endl ;

													// on tire les noeuds à exclure du graphe g1

			int nb_noeuds_Out_exclus = nb_out - get<1>(supports_observes.at(j)) ;
			int nb_noeuds_In_exclus  = nb_in  - get<2>(supports_observes.at(j)) ;

			Tvertice_identifiers_set* drawn_vertices = new Tvertice_identifiers_set ;


													// tire au hasard les noeuds
													// et les place dans drawn_vertices
													// ce sont les noeuds à exclure

			get_random_vertices2(nb_noeuds_Out_exclus, tVO, nb_out, *drawn_vertices) ;
			get_random_vertices2(nb_noeuds_In_exclus, tVI, nb_in, *drawn_vertices) ;

			// displaySet("drawn", *drawn_vertices) ;

													// on élimine les noeuds à exclure
			g1->discard_vertices(*drawn_vertices) ;

			delete (drawn_vertices) ;

			graph<string>* gA ;

			if (graphe_dirige)
			{
				directed_graph<string>* gp =  dynamic_cast<directed_graph<string>* > (g1) ;
				gA = new directed_graph<string>(*gp) ;
			}
			else
				gA = new graph<string>(*g1) ;

			Troles_set roles ;
			gA->apply_abstractions(a, 1, initial_graph, roles) ;	// abstraction avec min_sup = 1


			int nb_noeuds_abstraction = gA->nb_vertices() ;

			// cout << "   nb noeuds = " << nb_noeuds_abstraction << endl ;

													// suport abstrait attendu
			sommes_supports_attendus[j]        += nb_noeuds_abstraction ;
			sommes_carres_supports_attendus[j] += nb_noeuds_abstraction *
												  nb_noeuds_abstraction ;

			delete (gA) ;
			delete (g1) ;
		}
		// fin j (supports)

	}
	// fin i (tirages)
												// calcul des supports abstraits attendus
												// a partir des sommes
	for (int j = 0 ; j < nb_supports ; j++)
	{
		supports_attendus[j] = sommes_supports_attendus[j] / (n * 1.0) ;

		ecarts_types[j] = sqrt(
								(sommes_carres_supports_attendus[j] / (n *1.0)) -
								(supports_attendus[j] * supports_attendus[j])
							  ) ;
	}

	delete[] sommes_carres_supports_attendus ;
	delete[] sommes_supports_attendus ;
}


// ============================== MAIN ========================================

int main(int argc, char* argv[])
{
  vector<string> objects_names ;
  vector<vector<int> > initial_objects ;
  Tadjacences_map  adjacences ;
  vector<string> description_noeuds ;					// pas utilisé mais indispensable
  vector<tuple<string, int,int, int> > supports_observes ;
  vector<string> items_names ;
  vector<vector<int> > transacts ;
  Tweighted_edges_map edges_map ;

  bool local	  				= false ;		// local ou global
  bool communauty_abstraction 	= false ;		// abstraction dans les communautés ounon
  int min_sup	  				= 1     ;		// frequence minimal des patterns a considérer
												// pas de graphe des triangles
  abstraction a1 ;
  abstraction a2 ;
  string name_data_file       	= argv[1]   ;
  bool graphe_dirige			= false     ;	// dirige par defaut
  bool graphe_pondere           = false     ;
  
  string graph_data_file    = argv[1]       ;
  string supports_data_file = argv[2]       ;
  int  n  = 1                               ;   // nombre de repetitions

  set<Tvertice_identifier> VO ;							// noms noeuds ouptut
  set<Tvertice_identifier> VI ;							// noms noeuds input
  map<string, Tvertice_identifier> vertice_idents ;

  if (!handle_supports_attendus_parameters(argc, argv, a1, graphe_dirige, graphe_pondere, n))
	  return 0 ;

  write_supports_attendus_parameters(argv[0], graph_data_file, supports_data_file,
		                             a1, graphe_dirige, graphe_pondere, n) ;

  cerr << endl << "reading graph data..." << endl ;


 read_data_new_ri(graphe_dirige, graphe_pondere, name_data_file, objects_names,
		  	  	   items_names, transacts, adjacences, edges_map) ;


 detect_out_in(objects_names, VO, VI, vertice_idents) ;

 cerr << endl << "reading supports..." << endl ;

 read_supports_data(supports_data_file, VO, VI, vertice_idents, supports_observes) ;

  // displayVectorTuples("supports", supports_observes) ;
  sort(supports_observes.begin(), supports_observes.end(), sort_pred()) ;
  // displayVectorTuples("supports", supports_observes) ;

 directed_graph<string > dg ;

  // cerr << "create graphe initial" << endl ;


  graph<string>* g = create_initial_graph(graph_data_file, false, objects_names,
				 	 	 	 	 	 	  description_noeuds, adjacences, edges_map,
										  graphe_dirige, graphe_pondere, a1, a2) ;

  // g->display_graph() ;

  // cerr << "fin creer graphe initial " << endl ;

  // displayVector("objects_names", objects_names) ;
  // cout << endl ;

  int nb_supports = supports_observes.size() ;

  // cout << "nb_supports = " << nb_supports << endl ;

  double* supports_attendus = new double [nb_supports] ;
  double* ecarts_types      = new double [nb_supports] ;


  time_t debut ;
  time(&debut) ;

  try
  {
	  cerr << "computing supports..." << endl ;

	  calcul_supports_attendus_bi(g, graphe_dirige, VO, VI, supports_observes, n, a1,
			  	  	  	          supports_attendus, ecarts_types) ;
  	  time_t fin ;
  	  time(&fin) ;

  	  cout << "#" << endl ;
  	  cout << "# begin      = " << ctime(&debut) ;
  	  cout << "# end        = " << ctime(&fin)  ;
  	  cout << "# duration   = " << difftime(fin,debut) << " seconds." << endl ;
  }
  catch(exception& e)
  {
	  cerr << e.what() << endl ;
  }
  catch(string& s)
  {
	  cerr << s << endl ;
  }

  delete (g) ;

  sauvegarder_supports(supports_observes, supports_attendus, ecarts_types) ;

  delete[] ecarts_types ;
  delete[] supports_attendus ;


  cerr << "end of the program " << argv[0] << endl ;

  return 0 ;
}
