/**
 * SUPPORTS_ATTENDUS_LOCAUX.CPP
 * Dominique Bouthinon (LIPN, University Paris 13)
 *
 * THIS PROGRAM MUST COMPILED WITH VERSION C++11
 * WITH OPTIONS THAT ALLOW FOR <functionial>
 *
 */  
#include <cstdlib>
#include <algorithm>
#include <vector>
#include <sstream>
#include <fstream>
#include <iostream>
#include <sys/time.h>
#include <time.h>
#include <ctime>
#include <stdio.h>
#include <unistd.h>
#include <set>
#include <string>
#include <tuple>
#include <math.h>

#include "types_supports_attendus_locaux.hpp"
#include "read_write_data2.hpp"
#include "utils_divers.hpp"
#include "read_supports_locaux_input.hpp"
#include "common_functions.hpp"
#include "common_supports_functions.hpp"
#include "parameters.hpp"
#include "graph.hpp"
#include "directed_graph.hpp"

using std::cout;
using std::cerr;
using std::endl;

using namespace std;

	// chaque tuple observe est <c, |ext(c)|, |ext1(c)|, |ext(l)|, |ext1(l)|, |cc|, l)
	// où l est le clos de la composante connexe cc
void calcul_supports_attendus_locaux(
									   graph<string>* initial_graph,				// graphe initial
									   bool graphe_dirige,
									   vector<Type_tuple_observe>& supports_observes,	// (par ordre croissant) de |ext(c)|
									   int n,							// nombre d'iterations
									   abstraction a1,
									   abstraction a2,
									   bool a2_active,
									   int k_communauty,
									   float supports_attendus[],
									   float ecarts_types[]
			    	 	 	      )
{
													// chaque triplet de supports_observes contient :
													// element 0 : clos abstrait
													// element 1 : support non abstrait observé du clos
													// element 2 : support abstrait observé du clos

	srand(time(NULL)) ;									// initialisation séquence aleatoire

	int nb_supports = supports_observes.size() ;
													// met les identificateurs des noeuds de g
													// dans un tableau
	int nb_vertices = initial_graph -> nb_vertices() ;
	// cerr << "nb noeuds = " << nb_vertices << endl ;
	Tvertice_identifier vertices[nb_vertices] ;

	int id = 0 ;

	Tvertice_identifiers_set_iterator  it     = initial_graph->get_vertice_identifiers_list().begin()  ;
	Tvertice_identifiers_set_iterator  it_end = initial_graph->get_vertice_identifiers_list().end()  ;

	for ( ; it != it_end ; it++)
	{
		vertices[id] = *it ;
		// cout << vertices_g[id] << endl ;
		id++ ;
	}

	float sommes_supports_attendus[nb_supports] ;
	float sommes_carres_supports_attendus[nb_supports];

													// initialisation sommets
	for (int k = 0 ; k < nb_supports ; k++)
	{
		  sommes_supports_attendus[k] = 0 ;
		  sommes_carres_supports_attendus[k] =  0 ;
	}

														// faire n tirages
	for (int i = 0 ; i < n ; i++)					// i identifie un tirage
	{
		// cerr << "tirage  " << i << endl ;
													// copie graphe initial
	
		graph<string>* g1 = initial_graph->copy_graph() ;

		/*
   								// copie du graphe initial
		if (graphe_dirige)
		{
			directed_graph<string>* gp = dynamic_cast<directed_graph<string>* > (initial_graph) ;
			g1 = new directed_graph<string>(*gp) ;
		}
		else
			g1 = new graph<string>(*initial_graph) ;
		*/

		// g1-> display_vertice_identifiers() ;
		int limite_sup_tirage = nb_vertices  ;
				 									// pour chaque support observe (du plus grand au plus petit)
													// (j identifie un clos abstrait, et un support)
		for (int j = 0 ; j < nb_supports ; j++)
		{
			Type_tuple_observe& t = supports_observes.at(j) ;

			/*
			cerr << endl ;
			cerr << "motif_c             = " << get<0>(t) 	<< endl ;
			cerr << "support_c           = " << get<1>(t) 	<< endl ;
			cerr << "support_abstrait_c  = " << get<2>(t) 	<< endl ;
			cerr << "support_l           = " << get<3>(t)	<< endl ;
			cerr << "support_abstrait1_l = " << get<4>(t) 	<< endl ;
			cerr << "support_cc_l        = " << get<5>(t)	<< endl ;
			cerr << "motif_l             = " << get<6>(t) 	<< endl ;
			*/


													// on tire les noeuds à exclure
			int nb_noeuds_a_tirer = limite_sup_tirage  - get<1>(t) ;

			Tvertice_identifiers_set drawn_vertices ;

													// tire au hasard supports[j] noeuds
													// et les place dans drawn_vertices
													// ce sont les noeuds à exclure

			/*
			cout << "nb_vertices_g      = " << nb_vertices 		 << endl ;
			cout << "limite_sup_tirage  = " << limite_sup_tirage << endl ;
			cout << "nb_noeuds_a_tirer  = " << nb_noeuds_a_tirer << endl ;

			displayTab(vertices, nb_vertices) ;
			displayTab(vertices, limite_sup_tirage) ;
			*/


			get_random_vertices(nb_noeuds_a_tirer, vertices, limite_sup_tirage, drawn_vertices) ;

			/*
			cout << "limite_sup_tirage= " << limite_sup_tirage << endl ;
			cout << "apres tirage : " << endl ;
			displayTab(vertices, nb_vertices) ;
			displayTab(vertices, limite_sup_tirage) ;
			*/

			/*
			 graph<string>::display_vertice_identifiers_set(drawn_vertices) ;
			 cout << endl ;

			 cout << "noeuds g1 avant effacement " << endl ;
			 g1->display_vertice_identifiers() ;
			 */
													// on élimine les noeuds à exclure
			g1->discard_vertices(drawn_vertices) ;


			// cout << "noeuds g1 apres effacement " << endl ;
			// g1->display_vertice_identifiers() ;


			graph<string>* gA = g1->copy_graph() ;
												// copie du graphe initial
			/*
			if (graphe_dirige)
			{
				directed_graph<string>* gp =  dynamic_cast<directed_graph<string>* > (g1) ;
				gA = new directed_graph<string>(*gp) ;
			}
			else
				gA = new graph<string>(*g1) ;
			*/

			// gA->display_graph() ;
			// cout << endl << "ga (avant abstraction) nombre noeuds = " << gA->nb_vertices() << endl ;


			Troles_set roles ;
			gA->apply_abstractions(a1, 1, initial_graph,roles) ;	// abstraction avec min_sup = 1

			// gA->display_vertice_identifiers() ;

			Troles_set roles2 ;

			if (a2_active)
				gA ->apply_abstractions(a2,1, initial_graph, roles2) ;

			// gA->display_vertice_identifiers() ;

											// calcule des composantes connexes
			gA->compute_connected_components() ;

			// gA->display_connected_components() ;

											// calcule de la taille moyenne des
											// composantes connexes

			int sum_size_component = 0 ;
											// pour chaque composante connexe de g
			for (Tconnected_components_iterator it = gA->get_connected_components().begin() ;
			     it != gA->get_connected_components().end( ) ; it++)
			{
				   Tvertice_identifiers_set& component = (*it).second ;

				   // cerr << "component.size() = " << component.size() ;
				   sum_size_component += component.size() ;
			}

			float mean_size_component ;

			if (gA->get_connected_components().size() > 0)
				mean_size_component = (sum_size_component * 1.0) /
									  gA->get_connected_components().size() ;
			else
				mean_size_component = 0 ;

			// cerr << "mean_size_component =" << mean_size_component << endl ;

			// gA->display_graph() ;
			// cout << endl << "ga nombre noeuds = " << gA->nb_vertices() << endl ;

													// suport abstrait attendu
			sommes_supports_attendus[j]        += mean_size_component ;
			sommes_carres_supports_attendus[j] += mean_size_component *
												  mean_size_component ;

			// cout << "sommes_supports = " << sommes_supports_attendus[j] << endl ;
			// cout << "sommes_carres = "   << sommes_carres_supports_attendus[j] << endl ;

			delete (gA) ;
		}

		delete(g1) ;
	}
												// calcul des supports abstraits locaux attendus
												// a partir des sommes
	for (int j = 0 ; j < nb_supports ; j++)
	{
		/*
		cout << endl ;
		cout << "motif_c = " << get<0>(supports_observes.at(j)) << endl ;
		cout << "motif_l = " << get<6>(supports_observes.at(j)) << endl ;
		cerr << "support_cc_l  = " << get<5>(supports_observes.at(j))<< endl ;
		cout << "sommes_supports = " << sommes_supports_attendus[j] << endl ;
		cout << "sommes_carres = " << sommes_carres_supports_attendus[j] << endl ;
		*/

		supports_attendus[j] = sommes_supports_attendus[j] / (n * 1.0) ;
		ecarts_types[j] = sqrt(
								(sommes_carres_supports_attendus[j] / (n *1.0)) -
								(supports_attendus[j] * supports_attendus[j])
							  ) ;
		/*
		cout << "support abstrait observe = " << get<5>(supports_observes.at(j)) << endl ;
 		cout << "support abstrait attendu = " << supports_attendus[j] << endl ;
		cout << "ecart type = "      << ecarts_types[j] << endl ;
		*/
	}
}
									

void write_supports_attendus_locaux_parameters(string program,
					  string graph_data_file, string supports_data_file, bool local,
					  bool build_triangles_graph, abstraction& a1, abstraction& a2,
					  bool a2_active, int k_communauty, bool directed_graph, bool weighted_graph)
{
  cout << "# program                          = " << program		  	   << endl ;
  cout << "# graph data file                  = " << graph_data_file 	   << endl ;
  cout << "# observed supports file           = " << supports_data_file  << endl ;
  cout << "# local                            = " << local       	  	   << endl ;
  cout << "# build_triangles_graph            = " << build_triangles_graph	  << endl  ;
  cout << "# k-2 communauty                   = " << k_communauty << endl ;
  write_abstractions(a1, "(a1)") ;
  write_abstractions(a2, "(a2)") ;
  cout << "# a2_active                        = " << a2_active << endl ;
  cout << "# directed_graph                   = " << directed_graph	    << endl ;
  cout << "# weighted_graph                   = " << weighted_graph	    << endl ;
  cout << "#" << endl ;
}


			//	gestion des options du programme
bool handle_supports_attendus_locaux_parameters(
								   int argc,
								   char* argv[],
								   bool& build_triangles_graph,
								   abstraction& a1,
								   abstraction& a2,
								   bool& a2_active,			// temoigne que a2 differe de a1 et doit etre applique
								   int& k_communauty,
								   bool& directed_graph,
								   bool& weighted_graph,
								   int& n
					 				 )
{
	bool ok = true ;

	int i = 3 ;							// 3 premiers parametres vus avant

	while (i < argc && ok)
	{
		// cout << argv[i] << endl ;

		if(string(argv[i]) == "-tg")
		{
			build_triangles_graph = true ;
			i++ ;
		}
		else if(string(argv[i]) == "-n")
		{
					n = atoi(argv[i+1]) ;
					i = i + 2 ;
		}
		else if(string(argv[i]) == "-directed")
		{
			directed_graph = true ;
			i++ ;
		}
		else if(string(argv[i]) == "-weighted")
			{
				weighted_graph = true ;
				i++ ;
			}
		else if(string(argv[i]) == "-u")			// k communautes
		{
			k_communauty = atoi(argv[i+1]) ;

			if (k_communauty < 3)
			{
				cout << "!!! ERROR !!! " << endl ;
				cout << "-u " << k_communauty << endl ;
				cout << "the argument must be greater than 2" << endl ;
				ok = false ;
			}

			i = i + 2 ;
		}
		else if (handle_abstractions(argv, a1, i, "1")) 
		{
		}
		else if (handle_abstractions(argv, a2, i, "2")) 
		{
		}
		else
		{
			ok = false ;
			cerr << "parameter " << argv[i] << " not recognized" << endl ;
		}
	}

	// NOTE : k_dense == 3 <=> abstraction triangles

		// l'abstraction a2 doit aussi contenir a1,
		// sauf si a2 ne contient qu'une abstraction en
		// composantes connexes (l'abstraction a1 sera
		// toujour satisfaite dans a2).

		// (pour chaque abstraction (degre, cc, etc...)
		// il faut retenir pour a2 la plus contraignante entre a1 et a2

	if (ok)
		return set_a2_abstraction(a1, a2, directed_graph, a2_active) ;

	return false ;
}



void sauvegarder_supports(
		 	 	 	 	    vector<Type_tuple_observe>& supports_observes,
							float supports_attendus[],
							float ecarts_types[]
						 )
{
	int nb_supports = supports_observes.size() ;

	for (int i = 0 ; i < nb_supports ; i++)
	{
		Type_tuple_observe& t = supports_observes[i] ;

		cout << get<0>(t) << ", "
			 << get<1>(t) << ", "
			 << get<2>(t) << ", "
			 << get<3>(t) << ", "
			 << get<4>(t) << ", "
			 << get<5>(t) << ", "
			 << supports_attendus[i] << ", "
			 << ecarts_types[i] << ", "
			 << get<6>(t) << endl ;
	}
}

				// compare deux tuples selon son deuxième argument
struct sort_pred
{
    bool operator()(const Type_tuple_observe &left, const Type_tuple_observe &right)
    {
        return get<1>(left) > get<1>(right);
    }
} ;


// ============================== MAIN ========================================

int main(int argc, char* argv[])
{
  vector<string> objects_names ;
  vector<string> items_names ;
  vector<vector<int> > transacts ;

  vector<vector<int> > initial_objects ;
  Tadjacences_map  adjacences ;
  vector<string> description_noeuds ;	// pas utilise mais indispensable
  vector<Type_tuple_observe> supports_observes ;

  string graph_data_file    = argv[1]       ;
  string supports_data_file = argv[2]       ;
  int  n  = 1                               ;   					// nombre de tirages
  bool local = true                         ;
  bool build_triangles_graph = false        ;	// frequence minimal des patterns a considérer								
										// pas de graphe des triangles
  abstraction a1 ;
  abstraction a2 ;
  bool a2_active = false ;
  int k_communauty = 0 ;
  bool graphe_dirige  = false ;
  bool graphe_pondere = false ;
  Tweighted_edges_map edges_map ;
  string name_data_file       	= argv[1]   ;


  if (!handle_supports_attendus_locaux_parameters(argc, argv, build_triangles_graph, a1, a2, a2_active, 
                                                  k_communauty, graphe_dirige, graphe_pondere, n))
	return 0 ;

   write_supports_attendus_locaux_parameters(argv[0], graph_data_file, supports_data_file,
		            					local, build_triangles_graph, a1, a2,
	                					a2_active, k_communauty, graphe_dirige, graphe_pondere) ;

  cerr << endl << "reading graph data..." << endl ;


   read_data_new_ri(graphe_dirige, graphe_pondere, name_data_file, objects_names,
  		  	  	   items_names, transacts, adjacences, edges_map) ;


   // read_graph_data2(graphe_dirige, graph_data_file, objects_names, adjacences) ;

  // afficher_adjacences(adjacences) ;
  // cout << endl ;

  cerr << "reading support data..." << endl ;
  read_supports_data_local(supports_data_file, supports_observes) ;

  // displayVectorTuples("supports", supports_observes) ;
  sort(supports_observes.begin(), supports_observes.end(), sort_pred()) ;
  // displayVectorTuples("supports", supports_observes) ;

  graph<string>* initial_graph = create_initial_graph(graph_data_file, false,
										   objects_names, description_noeuds,
										   adjacences, edges_map,
								             graphe_dirige, graphe_pondere, a1, a2) ;

  // g->display_graph() ;

  // displayVector("objects_names", objects_names) ;
  // cout << endl ;

  int nb_supports = supports_observes.size() ;
  float supports_attendus[nb_supports] ;
  float ecarts_types[nb_supports] ;

  time_t debut ;
  time(&debut) ;

  try
  {
	  cerr << "computing supports..." << endl ;

	  calcul_supports_attendus_locaux(initial_graph, graphe_dirige, supports_observes, n, a1, a2, a2_active,
			  	  	  	  	    k_communauty, supports_attendus, ecarts_types) ;
  	  time_t fin ;
  	  time(&fin) ;

  	  cout << "#" << endl ;
  	  cout << "# begin      = " << ctime(&debut) ;
  	  cout << "# end        = " << ctime(&fin)  ;
  	  cout << "# duration   = " << difftime(fin,debut) << " seconds." << endl ;
  	  cout << endl ;
  }
  catch(exception& e)
  {
	  cerr << e.what() << endl ;
  }
  catch(string& s)
  {
	  cerr << s << endl ;
  }

  delete (initial_graph) ;

  sauvegarder_supports(supports_observes, supports_attendus, ecarts_types) ;

  cerr << "end of the program " << argv[0] << endl ;

  return 0 ;
}
